# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'color_picker_dialog_box.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from .constants import *

class Ui_Dialog(object):
    def setupUi(self, Dialog, label, color_btn):
        self.mask = display_settings["list_of_mask"][-1]
        self.label_target = label 
        self.color_btn = color_btn
        self.Dialog = Dialog
        self.Dialog.setObjectName("Dialog")
        self.Dialog.resize(381, 216)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(30, 170, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.plainTextEdit = QtWidgets.QPlainTextEdit(self.Dialog)
        self.plainTextEdit.setGeometry(QtCore.QRect(10, 60, 361, 31))
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.label = QtWidgets.QLabel(self.Dialog)
        self.label.setGeometry(QtCore.QRect(150, 30, 71, 16))
        self.label.setObjectName("label")
        self.changeColor = QtWidgets.QPushButton(self.Dialog)
        self.changeColor.setGeometry(QtCore.QRect(150, 120, 75, 23))
        self.changeColor.setObjectName("pushButton")
        self.changeColor.clicked.connect(self.openColorPicker)

        self.retranslateUi(self.Dialog)
        self.buttonBox.accepted.connect(self.confirm_label_and_exit) # type: ignore
        self.buttonBox.rejected.connect(self.Dialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(self.Dialog)

    def confirm_label_and_exit(self):
        # Get the text from the QPlainTextEdit
        text = self.plainTextEdit.toPlainText()
        if len(text)>0:
            # Set the text to the QLabel
            self.label_target.setText(text)
        self.Dialog.accept()

    def hex_to_rgb(self,hex_color):
        # Remove the hash '#' if present
        hex_color = hex_color.lstrip('#')

        # Convert the hex string to an integer tuple
        rgb_tuple = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))

        return rgb_tuple

    def get_previous_color(self, stylesheet:str):
        stylesheet = stylesheet.replace(" ","")
        stylesheet = stylesheet.split("background-color:rgb")[1]
        stylesheet = stylesheet.split(";")[0]
        colors = stylesheet[1:-1].split(",")
        colors = [int(c) for c in colors]
        # turn it into bgr, which is cv2 default
        colors = (colors[2],colors[1],colors[0])
        return colors

    def openColorPicker(self):
        # Open the color picker dialog
        color = QColorDialog.getColor()

        # Check if a valid color is selected
        if color.isValid():
            # Get the current stylesheet
            current_stylesheet = self.color_btn.styleSheet()
            
            # Remove any existing background-color entry
            if "background-color" in current_stylesheet:
                previous_color = self.get_previous_color(current_stylesheet)
                # Remove the existing background-color rule
                start_idx = current_stylesheet.index("background-color:")
                end_idx = current_stylesheet.index(";", start_idx)
                current_stylesheet = current_stylesheet[:start_idx] + current_stylesheet[end_idx + 1:].strip()

            # Add the new background color to the cleaned-up stylesheet
            # rgb is needed for the CSS stylesheet from the button colors
            new_rgb = self.hex_to_rgb(color.name())
            new_stylesheet = f"{current_stylesheet} background-color: rgb{new_rgb};"
            # bgr is needed to change the color of the brush to its new value 
            new_bgr = new_rgb[2],new_rgb[1],new_rgb[0]
            
            # set the new stylesheet for this button and update the pixels in the screen that matched the previous color
            self.color_btn.setStyleSheet(new_stylesheet)
            brush_settings["color"] = new_bgr
            
            # bgr_alpha is needed for the pixel values in the mask, as understood by cv2
            new_bgr_alpha = (*new_bgr,255)
            for mask in display_settings["list_of_mask"]:
                # replace the pixels with the previous color value for the new color in RGB format
                filtered_rgb_from_mask = mask[...,:3] #filter for all the values expect alpha channel
                mask_indices = (filtered_rgb_from_mask == previous_color).all(axis=-1)  # Find matching pixels
                mask[mask_indices] = new_bgr_alpha  # Update those pixels with the new color

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label.setText(_translate("Dialog", "Label Name"))
        self.changeColor.setText(_translate("Dialog", "Color"))


if __name__ == "__main__":
    import sys
    # Enable High DPI scaling
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    # Use the high-resolution icons and fonts
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'warning_delete.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.



from .constants import *


class Ui_Dialog(object):
    def setupUi(self, Dialog, idx_delete, lobutton_label_block, button):
        self.button = button #button is different than items inside lobuton_label_block list
        self.lobutton_label_block = lobutton_label_block
        self.idx_delete = idx_delete
        self.Dialog = Dialog
        self.Dialog.setObjectName("Dialog")
        self.Dialog.resize(400, 182)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(30, 130, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.label = QtWidgets.QLabel(self.Dialog)
        self.label.setGeometry(QtCore.QRect(20, 70, 341, 51))
        self.label.setObjectName("label")
        self.label_2 = QtWidgets.QLabel(self.Dialog)
        self.label_2.setGeometry(QtCore.QRect(120, 20, 221, 51))
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_2.setFont(font)
        self.label_2.setTextFormat(QtCore.Qt.PlainText)
        self.label_2.setScaledContents(False)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(self.Dialog)
        self.label_3.setGeometry(QtCore.QRect(140, 20, 41, 41))
        self.label_3.setText("")
        self.label_3.setPixmap(QtGui.QPixmap("icons/warning_icon.png"))
        self.label_3.setScaledContents(True)
        self.label_3.setObjectName("label_3")

        self.retranslateUi(self.Dialog)
        self.buttonBox.accepted.connect(self.delete_idx) # type: ignore
        self.buttonBox.rejected.connect(self.Dialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(self.Dialog)

    def get_previous_color(self, stylesheet:str):
        stylesheet = stylesheet.replace(" ","")
        stylesheet = stylesheet.split("background-color:rgb")[1]
        stylesheet = stylesheet.split(";")[0]
        colors = stylesheet[1:-1].split(",")
        colors = [int(c) for c in colors]
        # turn it into bgr, which is cv2 default
        colors = (colors[2],colors[1],colors[0])
        return colors

    def delete_idx(self):
        # First remove the button or label widget from the layout
        widget_to_remove = self.lobutton_label_block[self.idx_delete]
        
        # Assuming `widget_to_remove` is a widget object (like a QPushButton, QLabel, etc.)
        if isinstance(widget_to_remove, QtWidgets.QWidget):
            widget_to_remove.setParent(None)  # Detach from layout
            widget_to_remove.deleteLater()    # Marks the widget for deletion
        
       
        # update brush color to default
        brush_settings["color"] = (0,255,0)

        ######## DELETING THE COLOR IN THE MASK ##################
        current_stylesheet = self.button.styleSheet()
        print("#"*60,current_stylesheet)
        previous_color = self.get_previous_color(current_stylesheet)
        
        # bgr_alpha is needed for the pixel values in the mask, as understood by cv2
        new_bgr_alpha = (0,0,0,0)
        # replace the pixels with the previous color value for the new color in RGB format
        for mask in display_settings["list_of_mask"]:
            filtered_rgb_from_mask = mask[...,:3] #filter for all the values expect alpha channel
            mask_indices = (filtered_rgb_from_mask == previous_color).all(axis=-1)  # Find matching pixels
            mask[mask_indices] = new_bgr_alpha  # Update those pixels with the new color
        

        # Refresh the layout after the widget is removed
        self.Dialog.update()
        
        # Remove the element from the list
        self.lobutton_label_block.pop(self.idx_delete)
        # Close the dialog after the deletion
        self.Dialog.accept()

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label.setText(_translate("Dialog", "This action cannot be undone. Please be sure before clicking OK"))
        self.label_2.setText(_translate("Dialog", "Careful!"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



import numpy as np
from .constants import *
# DD. POINT
# pt = Coordinate(int,int)
# interp. a point in the scaled view of a webcam feed, scaled to fit the imageDisplay Widget of the main program
pt = Coordinate(0,0)

# DD. POLYGON
# polygon = {"DONE":False, "POINTS":[POINT, ...]}
# interp. a collection of POINTs to trace a polygon. Coordinates are given relative to the scaled 
# area of the mask represented in the imageDisplay Widget of the main program
polygon = {"DONE":False, "POINTS":[pt],"COLOR":None}


# DD. POLYGON_MANAGER
# polyman = PolyMan()
# interp. a set of attributes used to define the polygons that will be created on top of the mask
class Polyman():
    def __init__(self):
        # self.firstClick = True #to start a new polyong
        self.current_polygon = {"DONE":False, "POINTS":[], "COLOR":None}
        # self.lopolygon = [self.current_polygon]
        
    def updatePoly(self, coor):
        coor_x, coor_y = coor
        self.current_polygon["POINTS"].append(Coordinate(coor_x,coor_y))
        
    def finishPolygon(self):
        if not brush_settings["is_brush_mode"] == "polygon":
            return
            
        points = np.array([[p.x, p.y] for p in self.current_polygon["POINTS"]], np.int32)
        points = points.reshape((-1, 1, 2))
        
        # evaluate if there are any points to annotate
        if len(points) == 0:
            return 
        
        if os_settings["substractive_mode"]:
            _boolean_mask = np.zeros_like(display_settings["list_of_mask"][-1])
            cv2.fillPoly(_boolean_mask, [points], color=(255,255,255))
            inverted_mask = cv2.bitwise_not(_boolean_mask)
            
            if os_settings["top_layer_edit"]:
                display_settings["list_of_mask"][-1] = cv2.bitwise_and(display_settings["list_of_mask"][-1], inverted_mask)
            else:
                for idx, mask in enumerate(display_settings["list_of_mask"]):
                    display_settings["list_of_mask"][idx] = cv2.bitwise_and(mask, inverted_mask)
        else:
            # Create a temporary mask to draw the polygon
            temp_mask = np.zeros_like(display_settings["list_of_mask"][-1])
            cv2.fillPoly(temp_mask, [points], color=(255, 255, 255, 255))  # Fill with white
            
            # Create a binary mask of where the polygon was drawn (ignoring alpha channel)
            binary_mask = np.any(temp_mask[:, :, :3] > 0, axis=2)
            
            # Create a new mask with the desired color
            _mask = add_empty_mask()
            _mask[binary_mask] = brush_settings["color"]
            
            # Append the new mask
            # display_settings["list_of_mask"].append(new_mask)
        
        # Reset the current polygon
        self.current_polygon = {"DONE": False, "POINTS": [], "COLOR": None}


    def pop_last_point(self):
        if len(self.current_polygon["POINTS"])>0:
            self.current_polygon["POINTS"].pop(-1)


import numpy as np
from .constants import *

# DD. POINT
# pt = Coordinate(int,int)
# interp. a point in the scaled view of a webcam feed, scaled to fit the imageDisplay Widget of the main program
pt = Coordinate(0,0)

# DD. BINARY_MASK_MANAGER
# binman = Binman()
# interp. a set of components that define the behavior of the binary mask tool
class Binman():
    def __init__(self, parent=None):
        # Store a reference to the parent ImageAnnotator
        self.parent = parent
        self.threshold = brush_settings["binary_mask_thres"]  # default threshold value
        self.binary_mask = None

    def onMouseEventDown(self):
        """
        Implements a binary mask tool that creates a new mask based on the annotation upon click event
        """
        
        
        # Create a copy of the current image 
        thres_target = display_settings["image"].copy()
        
        # Convert image to grayscale
        thres = cv2.cvtColor(thres_target, cv2.COLOR_BGR2GRAY)
        
        # Apply threshold to create binary mask
        _, self.binary_mask = cv2.threshold(thres, self.threshold, 255, cv2.THRESH_BINARY)
        # If subtractive mode is enabled, invert the mask
        if os_settings["substractive_mode"]:
            self.binary_mask = cv2.bitwise_not(self.binary_mask)

        # Convert single-channel mask to 3-channel
        self.binary_mask = cv2.merge([self.binary_mask] * 4)  # Expands grayscale mask to (H, W, 3)
        
    # # Original working version v1
    # def finish_binary_mask(self):
    #     if self.binary_mask is None:
    #         return

    #     display_settings["list_of_mask"].append(np.zeros_like(display_settings["list_of_mask"][-1]))

    #     # Convert the RGB part to grayscale to detect the drawn region
    #     binary_mask_gray = cv2.cvtColor(self.binary_mask[:, :, :3], cv2.COLOR_RGB2GRAY)
    #     _, thresh = cv2.threshold(binary_mask_gray, 1, 255, cv2.THRESH_BINARY)

    #     # Find contours from the mask
    #     contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    #     # Fill the new mask directly
    #     cv2.fillPoly(display_settings["list_of_mask"][-1], contours, color=brush_settings["color"])

    
    # Working version with enhanced gap visibility
    def finish_binary_mask(self):
        if self.binary_mask is None:
            return
        _new_mask = add_empty_mask()
        gray_mask = cv2.cvtColor(self.binary_mask[:, :, :3], cv2.COLOR_RGB2GRAY)
        _, binary_mask = cv2.threshold(gray_mask, 1, 255, cv2.THRESH_BINARY)

        # Apply opening instead of closing to emphasize gaps
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))  # try (5,5) if needed
        opened_mask = cv2.morphologyEx(binary_mask, cv2.MORPH_OPEN, kernel)

        contours, hierarchy = cv2.findContours(opened_mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)

        cv2.fillPoly(_new_mask, contours, color=brush_settings["color"])
        # display_settings["list_of_mask"].append(new_mask)
        self.binary_mask = None

    


    def display(self, image):
        if self.binary_mask is not None:
            return self.binary_mask
        return image


import numpy as np
import math
from .constants import *

# DD. FREEHAND_MANAGER
# freehandman = FreehandMan()
# interp. a set of attributes used to define the freehand polygons that will be created on top of the mask
class Freehandman():
    def __init__(self):
        self.current_polygon = {"DONE": False, "POINTS": [], "COLOR": None}
        self.distance_threshold = 4  # Minimum distance to add a new point
        self.started = False
        
    def updatePolyOnMove(self, coor):
        """
        Updates the polygon based on cursor movement.
        Adds a new point only if the distance from the last point is greater than threshold.
        """
        if not self.started:
            return

        coor_x, coor_y = coor
        
        # If this is the first point, add it directly
        if len(self.current_polygon["POINTS"]) == 0:
            self.current_polygon["POINTS"].append(Coordinate(coor_x, coor_y))
            return
        
        # Get the last point
        last_point = self.current_polygon["POINTS"][-1]
        
        # Calculate Pythagorean distance
        distance = math.sqrt((coor_x - last_point.x)**2 + (coor_y - last_point.y)**2)
        
        # Only add point if distance is greater than threshold
        if distance > self.distance_threshold:
            self.current_polygon["POINTS"].append(Coordinate(coor_x, coor_y))
    
    def startFreehand(self, coor):
        """
        Starts a new freehand polygon at the given coordinate.
        Call this when the user starts drawing (e.g., mouse down).
        """
        coor_x, coor_y = coor
        self.current_polygon = {"DONE": False, "POINTS": [], "COLOR": None}
        self.current_polygon["POINTS"].append(Coordinate(coor_x, coor_y))
        self.started = True
        
    def finishPolygon(self):
        """
        Finishes the current polygon and applies it to the mask.
        Call this when the user presses Enter.
        """
        if not brush_settings["is_brush_mode"] == "freehand":  # Updated condition
            return
            
        points = np.array([[p.x, p.y] for p in self.current_polygon["POINTS"]], np.int32)
        points = points.reshape((-1, 1, 2))
        
        # evaluate if there are any points to annotate
        if len(points) == 0:
            return 
        
        if os_settings["substractive_mode"]:
            _boolean_mask = np.zeros_like(display_settings["list_of_mask"][-1])
            cv2.fillPoly(_boolean_mask, [points], color=(255, 255, 255))
            inverted_mask = cv2.bitwise_not(_boolean_mask)
            
            if os_settings["top_layer_edit"]:
                display_settings["list_of_mask"][-1] = cv2.bitwise_and(display_settings["list_of_mask"][-1], inverted_mask)
            else:
                for idx, mask in enumerate(display_settings["list_of_mask"]):
                    display_settings["list_of_mask"][idx] = cv2.bitwise_and(mask, inverted_mask)
        else:
            # Create a temporary mask to draw the polygon
            temp_mask = np.zeros_like(display_settings["list_of_mask"][-1])
            cv2.fillPoly(temp_mask, [points], color=(255, 255, 255, 255))  # Fill with white
            
            # Create a binary mask of where the polygon was drawn (ignoring alpha channel)
            binary_mask = np.any(temp_mask[:, :, :3] > 0, axis=2)
            
            # Create a new mask with the desired color
            _mask = add_empty_mask()
            _mask[binary_mask] = brush_settings["color"]
            
            # Append the new mask
            # display_settings["list_of_mask"].append(new_mask)
        
        # Reset the current polygon
        self.current_polygon = {"DONE": False, "POINTS": [], "COLOR": None}
        self.started = False

    def pop_last_point(self):
        """
        Removes the last point from the current polygon.
        """
        if len(self.current_polygon["POINTS"]) > 0:
            self.current_polygon["POINTS"].pop(-1)
            
    def get_current_points(self):
        """
        Returns the current polygon points for preview/rendering purposes.
        """
        return self.current_polygon["POINTS"]
    
    def clear_current_polygon(self):
        """
        Clears the current polygon without finishing it.
        """
        self.current_polygon = {"DONE": False, "POINTS": [], "COLOR": None}


from .constants import *

# CD. VideoDeviceManager()
# purp. manage the video device (webcam) and its properties
class VideoDeviceManager():
    def __init__(self):
        # Determine the width and height ratio of the camera
        self.webcam_idx = 0 #the initial index of the webcam
        self.get_avail_devices()
        self.read_new_VideoDevice(self.webcam_idx)
        self.in_pause = False # Whether to return empty frames or not
        self.image_captured_pause = False # Whether to return empty frames or not
        
    def get_avail_devices(self):
    #     print("""      
    #           Finding video devices available...
    #   .-------------------.
    #  /--_--.------.------/|
    #  |     |__||__| [==] ||
    #  |     | .--. | '''' ||
    #  |     || () ||      ||
    #  |     | `--' |      |/
    #  `-----'------'------'  Art by Joan Stark
    #  """)
        print("""
    ##     #####################     
    #      #######     ##########   ##
###     ##########         ###### #### 
           ####### LIVE PYXEL ### #### 
    ##     #######         ###### #### 
           #######     ##########   ##
 ##     ########################   

              """)
        available_cameras = []
        for device_index in range(4):
            cap = cv2.VideoCapture(device_index)
            if cap.isOpened():
                available_cameras.append(device_index)
                cap.release()  # Release the camera once checked
        self.avail_devices = available_cameras

    def release_video_device(self):
        '''
        Triggered when user ends the application, it terminates and releases all cameras that were opened.
        '''
        if hasattr(self, 'cap') and self.cap.isOpened():
            self.cap.release()
            self.cap = None
    
    def change_video_device(self, image_annotator):
        if bool(len(self.avail_devices)):
            self.webcam_idx = (self.webcam_idx + 1)%len(self.avail_devices)
        self.read_new_VideoDevice(self.webcam_idx)
        # OpenCV image processing setup
        image_annotator.image = self.get_image_from_webcam()  # Load first frame from webcam
        # self.image = cv2.resize(self.image, (W, H))  # Resize image to fit the window
        image_annotator.maskImage = np.zeros_like(image_annotator.image)  # Mask for drawing
            
    def read_new_VideoDevice(self,idx):
        global W, H
        self.cap = cv2.VideoCapture(idx)
        
        # Get the width and height of the frame
        W = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        H = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        print(f"Webcam resolution: {W}x{H}")


    # FD. get_image_from_webcam()
    # purp. create a surface class object with the feed from the webcam
    def get_image_from_webcam(self):
        _inf,frame = self.cap.read()
        if self.in_pause or self.image_captured_pause:
            return np.zeros((480, 720, 4), dtype=np.uint8)  # Return a blank image if frame is None
        surface = cv2.cvtColor(frame,0)
        # surface = cv2.rotate(surface, cv2.ROTATE_90_COUNTERCLOCKWISE)
        # surface = cv2.flip(surface,0)
        surface = cv2.resize(surface, (720,480))  #HEIGHT AND WIDTH get flipped because or the counterclockwise rotation
        return surface

    def save_image(self):
        final_mask = merge_masks(display_settings["list_of_mask"])
        final_mask[:, :, 3] = 255

        if not os_settings["images_path"]:
            QtWidgets.QMessageBox.critical(None, "Annotation aborted!", "A directory path couldn't be determined. Please create or load a project and try again")
        else:
            if np.max(final_mask[:, :, 0:3]) != 0:
                # Get the list of existing image files in the directory
                image_files = [f for f in os.listdir(os_settings["images_path"]) if f.startswith("img_") and f.endswith(".png")]
                
                # Find the highest numbered image
                max_num = 0
                for image_file in image_files:
                    try:
                        num = int(image_file.split("_")[1].split(".")[0])
                        if num > max_num:
                            max_num = num
                    except ValueError:
                        continue
                
                # Increment the number for the new image and mask
                new_num = max_num + 1
                image_name = f"img_{new_num:06d}.png"
                mask_name = f"msk_{new_num:06d}.png"
                
                # Save the new image and mask
                image_path = jn(os_settings["images_path"], image_name)
                mask_path = jn(os_settings["masks_path"], mask_name)
                cv2.imwrite(image_path, display_settings["image"])
                cv2.imwrite(mask_path, final_mask)
                
                # Restart the process with a new mask making the first element in the mask set
                _new_mask0 = np.zeros_like(self.get_image_from_webcam())
                display_settings["list_of_mask"] = [_new_mask0]
                
            else:
                QtWidgets.QMessageBox.critical(None, "Annotation aborted!", "Please draw a mask before saving the image")

class ImageEditor():
    '''
    Class to manage an image editor. Used to load images and masks from an already existing directory, throws a warning 
    if the directory is not set. The class also manages the image and mask lists, and the index of the current image.
    '''
    def __init__(self):
        self.index_image = 0
        # Storing the actual arrays avoids having to read the image and mask from the disk every time
        self.current_image = None 
        self.current_mask = None
        self.images_loaded = False # Refreshed when updating index, it flags if an image at index_image has been loaded
        # A list of dictionaries that store an image and a mask
        self.get_files()
    
    def next_image(self):
        self.index_image = (self.index_image + 1) % len(self.files)
        self.images_loaded = False
        self.get_files()
        display_settings["list_of_mask"] = [np.zeros_like(display_settings["list_of_mask"][-1])]
        self.get_image_from_folder()
    
    def prev_image(self):
        self.index_image = (self.index_image - 1) % len(self.files)
        self.images_loaded = False
        self.requires_resize = False
        self.original_height = None 
        self.original_width = None
        self.get_files()
        display_settings["list_of_mask"] = [np.zeros_like(display_settings["list_of_mask"][-1])]
        self.get_image_from_folder()
    
    
    def get_image_from_folder(self):
        '''
        Get the image from the folder specified by the user. The image is then resized to fit the window.
        '''
        if not os_settings["images_path"]:
            QtWidgets.QMessageBox.critical(None, "Annotation aborted!", 
                "A directory path couldn't be determined. Please create or load a project and try again\n"
                "You may have encountered a bug. Please contact author at garcilau@mcmaster.ca")
            return None
        
        if not self.images_loaded:
            try:
                # Read the image from the specified path
                _target_file = self.files[self.index_image]
                frame = cv2.imread(_target_file["imagePath"])
                if frame.ndim == 2:  # If grayscale
                    frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGRA)
                elif frame.shape[2] == 3:  # If BGR
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2BGRA)
                
                if frame is None:
                    raise ValueError(f"Could not read image from path: {_target_file['imagePath']}")
                
                # Get the expected dimensions from the first frame
                expected_height, expected_width = display_settings["image"].shape[:2]
                self.original_height, self.original_width = frame.shape[:2]

                
                self.requires_resize = False
                # Resize image (frame) if dimensions don't match
                if frame.shape[:2] != (expected_height, expected_width):
                    # print(f"Resizing mask from {frame.shape} to {(expected_height, expected_width)}")
                    frame = cv2.resize(frame, (expected_width, expected_height))
                    self.requires_resize = True
                
                # Clear and initialize the mask list
                # display_settings["list_of_mask"] = [zero_mask.copy()]
                
                # Load existing mask if available
                if os.path.exists(_target_file["maskPath"]):
                    _mask = cv2.imread(_target_file["maskPath"], cv2.IMREAD_UNCHANGED)  # Preserve original channels
                    if _mask.ndim == 2:  # If grayscale
                        _mask = cv2.cvtColor(_mask, cv2.COLOR_GRAY2BGRA)
                    elif _mask.shape[2] == 3:  # If BGR
                        _mask = cv2.cvtColor(_mask, cv2.COLOR_BGR2BGRA)
                    
                    # Resize mask if dimensions don't match
                    if _mask.shape[:2] != (expected_height, expected_width):
                        # print(f"Resizing mask from {_mask.shape} to {(expected_height, expected_width)}")
                        _mask = cv2.resize(_mask, (expected_width, expected_height))
                    
                    # Replace the zero mask with the loaded mask
                    display_settings["list_of_mask"][-1] = _mask
                else:
                    zero_mask = np.zeros((expected_height, expected_width, 4), dtype=np.uint8)  # 4-channel
                    display_settings["list_of_mask"][-1] = zero_mask.copy()

                self.images_loaded = True
                self.current_image = frame.copy()
                # self.current_mask = display_settings["list_of_mask"][-1].copy()
                return cv2.cvtColor(frame, 0)
            
            except Exception as e:
                QtWidgets.QMessageBox.critical(None, "Error loading image", 
                    f"An error occurred while loading the image:\n{str(e)}")
                return None
        else:
            # If the image is already loaded, just return it
            return self.current_image
    
    def save_image(self):
        '''
        Take the current image and mask. Save only the new mask under the same name as it previously had
        '''
        # statusBar = display_settings["statusBar"]
        final_mask = merge_masks(display_settings["list_of_mask"])
        final_mask[:, :, 3] = 255

        if not os_settings["images_path"]:
            QtWidgets.QMessageBox.critical(None, "Annotation aborted!", "A directory path couldn't be determined. Please create or load a project and try again")
        else:
            if np.max(final_mask[:, :, 0:3]) != 0:
                # # Get the list of existing image files in the directory
                # image_files = [f for f in os.listdir(os_settings["images_path"]) if f.startswith("img_") and f.endswith(".png")]
                
                # # Find the highest numbered image
                # max_num = 0
                # for image_file in image_files:
                #     try:
                #         num = int(image_file.split("_")[1].split(".")[0])
                #         if num > max_num:
                #             max_num = num
                #     except ValueError:
                #         continue
                
                # Increment the number for the new image and mask
                # new_num = max_num + 1
                # image_name = f"img_{new_num:06d}.png"
                # mask_name = f"msk_{new_num:06d}.png"
                
                # Save the new image and mask
                image_path = self.files[self.index_image]["imagePath"] 
                mask_path = self.files[self.index_image]["maskPath"]
                if self.requires_resize:
                    final_img = cv2.resize(display_settings["image"], (self.original_width, self.original_height))
                    final_mask = cv2.resize(final_mask, (self.original_width, self.original_height))
                else:
                    final_img = display_settings["image"]
                    final_mask = final_mask
                cv2.imwrite(image_path, final_img)
                cv2.imwrite(mask_path, final_mask)
                
                # Restart the process with a new mask making the first element in the mask set
                # new_zero_mask = np.zeros_like(display_settings["list_of_mask"][-1])
                # display_settings["list_of_mask"] = []
                # display_settings["list_of_mask"].append(new_zero_mask)
                # Show success message
                
                # Notify the user the image was successfully updated
                success_msg = f"Successfully saved mask to {os.path.basename(mask_path)}"
                show_status_message(success_msg, is_success=True)
                
            else:
                QtWidgets.QMessageBox.critical(None, "Annotation aborted!", "Please draw a mask before saving the image")

    
    def get_files(self):
        '''
        Use the directory specified by the user to gather all the images and masks, and turn them into a list of dictionaries.
        input: None
        output: None
        '''
        self.files = []
        if not os_settings["images_path"]:
            QtWidgets.QMessageBox.critical(None, "Process aborted!", "A directory path couldn't be determined. You may have encountered a bug. Please contact the author at garcilau@mcmaster.ca")
        else:
            # Get the list of image files in the directory
            image_files = [f for f in os.listdir(os_settings["images_path"]) if f.endswith(".png")]
            
            # Create a list of dictionaries with image and mask paths
            for image_file in image_files:
                mask_file = image_file.replace("img_", "msk_")
                # if nothing was changed, add msk_ to the name of the image
                mask_file = "msk_" + image_file if "msk_" not in mask_file else mask_file
                self.files.append({
                    "imagePath": jn(os_settings["images_path"], image_file),
                    "maskPath": jn(os_settings["masks_path"], mask_file)
                })
    
    def set_current_image_index(self,index):
        '''
        Set the current index of the image and mask to be loaded.
        input: index (int) - the index of the image and mask to be loaded
        output: None
        '''
        self.index_image = index
        self.images_loaded = False
        self.get_files()
        # find the maximum index of the files list
        max_index = len(self.files) - 1
        # if index is greater than the maximum index, set it to the maximum index
        if self.index_image > max_index:
            self.index_image = max_index
        self.get_image_from_folder()
        
        return self.index_image
# if __name__ == "__main__":
#     get_image_from_webcam()


from .constants import *  
from .video_device_manager import VideoDeviceManager, ImageEditor
from .ui_main_window import Ui_MainWindow
from .polyman import Polyman
from .bezierman import Bezierman
from .bucketman import Bucketman
from .binaman import Binman
from .freehandman import Freehandman
import platform, subprocess
# C.D. IMAGE_ANNOTATOR
# This is the main class that handles the image annotation tool. It is responsible for the following:
# 1. Handling the annotation tools (brush, polygon, bezier)
# 2. Handling the mouse events for drawing 
# 3. Handling the webcam integration >>> VideoDeviceManager()
# 4. Handling the image display >>> VideoDeviceManager()
# 5. Handling the UI components >>> Ui_MainWindow()
class ImageAnnotator(QMainWindow):
    def __init__(self):
        super().__init__()
        display_settings["statusBar"] = self.statusBar
        display_settings["statusBar"]().showMessage("ready")  # Initial message
        # Tools init
        self.polygon_manager = None
        self.bezier_manager = None
        self.bezier_current_pos = None
        self.freehand_current_pos = None
        self.binary_mask = None
        self.mainUI = Ui_MainWindow()
        self.bucket_manager = None
        self.mainUI.setupUi(self)
        self.imageFrozen = False #image captured

        # Webcam init
        self.deviceManager = VideoDeviceManager()
        cursor_settings["in_display"] = False #whether the cursor is inside the image display, which is the only time painting tools should work        

        # Image annotation init starts None because needs to know which folder to load
        self.editorManager = None
        # OpenCV image processing setup
        
        display_settings["image"] = self.deviceManager.get_image_from_webcam()
        display_settings["list_of_mask"] = [np.zeros_like(display_settings["image"])]
        
        # initialize the brush settings
        self.brush_preview_color = (0, 150, 0)  # Darker green for preview
        self.mainUI.brush_slider.setValue(70)
        self.mainUI.binarythres_slider.setValue(127) #min 0 and max 255
        brush_settings["opacity"] = self.mainUI.brush_slider.value()
        brush_settings["binary_mask_thres"] = self.mainUI.binarythres_slider.value()
        

        # Button webcam
        self.mainUI.btn_switch_webcam.clicked.connect(lambda: self.deviceManager.change_video_device(self))
        self.mainUI.brush_slider.valueChanged.connect(self.updateSliderBrushStrength)
        self.mainUI.binarythres_slider.valueChanged.connect(self.updateSliderBinaryThreshold)
        self.mainUI.btn_annotate.clicked.connect(self.save_image_handler)
        self.mainUI.btn_add.clicked.connect(lambda: self.mainUI.add_btn_block(label_text="New Label"))
        self.mainUI.btn_capture.clicked.connect(self.captureManager)
        self.mainUI.polygon_button.clicked.connect(lambda: self.update_paint_mode("polygon"))
        self.mainUI.brush_button.clicked.connect(lambda: self.update_paint_mode("brush"))
        self.mainUI.bezier_button.clicked.connect(lambda: self.update_paint_mode("bezier"))
        self.mainUI.bucket_button.clicked.connect(lambda: self.update_paint_mode("bucket"))
        self.mainUI.binmask_button.clicked.connect(lambda: self.update_paint_mode("binmask"))
        self.mainUI.freehand_button.clicked.connect(lambda: self.update_paint_mode("freehand"))
        self.mainUI.isAdditive_button.clicked.connect(self.update_isAdditive_mode)
        self.mainUI.isTopLayerOnly_button.clicked.connect(self.update_isTopLayerOnly_button_mode)
        self.mainUI.openFolder_button.clicked.connect(self.openFolder)
        self.mainUI.editMode_button.clicked.connect(self.openFolderForAnnotations)
        self.mainUI.next_btn.clicked.connect(lambda: self.updateImageInEditorMode("next", self.mainUI.number_input))
        self.mainUI.prev_btn.clicked.connect(lambda: self.updateImageInEditorMode("prev", self.mainUI.number_input))
        self.mainUI.number_input.returnPressed.connect(lambda: self.number_input_handler(self.mainUI.number_input))


        # Enable mouse tracking for the window and image display label
        self.setMouseTracking(True)
        self.mainUI.centralwidget.setMouseTracking(True)
        self.mainUI.imageDisplay.setMouseTracking(True)

        # Mouse tracking and drawing setup
        self.drawing = False  # Track if drawing
        self.last_point = QPoint()  # Track the last mouse point
        self.current_pos = QPoint()  # Track current mouse position for brush preview

        # Timer for updating webcam feed
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_webcam_feed)
        self.timer.start(30)  # Update every 30 milliseconds (~33 FPS)

        # Create a container for the edit button blocks and add the first buttonblock
        self.mainUI.add_btn_block(label_text="New Label")
        
        # Add Ctrl+Z behaviour
        # Create a shortcut for Ctrl+Z and connect it to a custom function
        self.shortcut_undo = QShortcut(QKeySequence("Ctrl+Z"), self)
        self.shortcut_undo.activated.connect(self.handle_undo)
        
        # Add ESC behaviour
        # Create a shortcut for Escape and connect it to a custom function
        self.shortcut_undo = QShortcut(QKeySequence("Escape"), self)
        self.shortcut_undo.activated.connect(self.reset_tool)
        
        # show application in fullscreen
        # self.showFullScreen()
        self.showMaximized()

        self.update_image_display()
        
    def number_input_handler(self, number_input):
        # Get the text from the input
        text = number_input.text()
        
        # Convert to integer (since we're using QIntValidator)
        if text:  # Check if the input is not empty
            try:
                number = int(text)
                if os_settings["config"] != "" and self.editorManager is not None:
                    update_text = self.editorManager.set_current_image_index(number)  # text updated after clamping to max index
                    number_input.setText(str(update_text))  # Update the input field with the clamped value
                # self.update_index_image(number)  # Call your helper method
            except ValueError:
                # This shouldn't happen because of QIntValidator, but just in case
                pass 
        
    def save_image_handler(self):
        '''
        Handle the right saving method depending upon the mode of the application
        '''
        if os_settings["webcam_mode"]:
            self.deviceManager.save_image()
        else:
            self.editorManager.save_image()
        
            
            
    def updateImageInEditorMode(self, direction, number_input):
        '''
        Update the image in editor mode
        '''
        if self.editorManager is not None:
            if direction == "next":
                self.editorManager.next_image()
                number_input.setText(str(self.editorManager.index_image))  # Update the input field with the current image index
            elif direction == "prev":
                self.editorManager.prev_image()
                number_input.setText(str(self.editorManager.index_image))  # Update the input field with the current image index
            
            # # Update the image display with the new image
            # self.update_image_display()
        
    def openFolderForAnnotations(self):
        '''
        Access the images and files for annotation
        '''
        # There are two main cases: User followed LivePyxel structure or has only images folder
        if os_settings["config"] != "":
            if os_settings["webcam_mode"]:                
                # path = os.path.normpath(os_settings["images_path"])  # Normalize the path format
                os_settings["webcam_mode"] = False
                # Pause the webcam feed
                self.deviceManager.in_pause = True
                # If there no ImageEditor instance, create one, which also automates finding the files
                if self.editorManager is None and os_settings["config"] != "":
                    self.editorManager = ImageEditor()
            else:
                # Reset the masks
                _msk = np.zeros_like(display_settings["list_of_mask"][-1])
                display_settings["list_of_mask"] = [_msk]
                # self.deviceManager.read_new_VideoDevice(0)  # Read the new video device
                self.deviceManager.in_pause = False
                os_settings["webcam_mode"] = True
                self.editorManager = None
        else:
            QMessageBox.warning(self, "Warning", "No directory has been set yet. Please open or create a new project")

    # Open the folder containing the images and masks
    def openFolder(self):
        """
        Open a folder in the default file explorer
        Works across Windows, macOS, and Linux
        """
        
        if os_settings["config"] != "":
            path = os.path.normpath(os_settings["images_path"])  # Normalize the path format
            if platform.system() == "Windows":
                os.startfile(path)
            elif platform.system() == "Darwin":  # macOS
                subprocess.run(["open", path])
            else:  # Linux
                subprocess.run(["xdg-open", path])
        else:
            QMessageBox.warning(self, "Warning", "No directory has been set yet. Please open or create a new project")

    # Change the status of the isAdditive button
    def update_isAdditive_mode(self):
        os_settings["substractive_mode"] = not os_settings["substractive_mode"]
        if os_settings["substractive_mode"]:
            self.mainUI.isAdditive_button.setIcon(self.mainUI.isAdditive_button_off_icon)
            
            brush_settings["color_before_substracting_mode"] = tuple(brush_settings["color"])
            if brush_settings["is_brush_mode"] != "binmask":
                brush_settings["color"] = (255, 255, 255)
            brush_settings["thickness"] = 2
        else:
            self.mainUI.isAdditive_button.setIcon(self.mainUI.isAdditive_button_on_icon)
            brush_settings["color"] = tuple(brush_settings["color_before_substracting_mode"])
            brush_settings["thickness"] = 1
    
    # Change the status of the isAdditive button
    def update_isTopLayerOnly_button_mode(self):
        os_settings["top_layer_edit"] = not os_settings["top_layer_edit"]
        if os_settings["top_layer_edit"]:
            self.mainUI.isTopLayerOnly_button.setIcon(self.mainUI.isTopLayerOnly_button_off_icon)
        else:
            self.mainUI.isTopLayerOnly_button.setIcon(self.mainUI.isTopLayerOnly_button_on_icon)
                

    def reset_tool(self):
        # if polygon mode
        if brush_settings["is_brush_mode"] == "polygon":
            self.polygon_manager = Polyman()
        elif brush_settings["is_brush_mode"] == "bezier":
            self.bezier_manager = Bezierman()
        elif brush_settings["is_brush_mode"] == "freeman":
            self.freehand_manager = Freehandman()
        

    def handle_undo(self):
        # if polygon mode
        if brush_settings["is_brush_mode"] == "polygon":
            self.polygon_manager.pop_last_point()
        elif brush_settings["is_brush_mode"] == "bezier":
            self.bezier_manager.pop_last_point()
        elif brush_settings["is_brush_mode"] == "freehand":
            self.freehand_manager.pop_last_point()

    # Update the tool MODE
    def update_paint_mode(self,mode:str):
        brush_settings["is_brush_mode"] = mode
        if brush_settings["is_brush_mode"] == "polygon":
            self.polygon_manager = Polyman()
        elif brush_settings["is_brush_mode"] == "freehand":
            self.freehand_manager = Freehandman()
        elif brush_settings["is_brush_mode"] == "bezier":
            self.bezier_manager = Bezierman()
        elif brush_settings["is_brush_mode"] == "bucket":
            self.bucket_manager = Bucketman(self)
        elif brush_settings["is_brush_mode"] == "binmask":
            self.binary_mask = Binman(self)
        elif brush_settings["is_brush_mode"] == "brush":
            _ = add_empty_mask()
        # iterate over the masks and erase any mask where the maximum value of RGB (without considering alpha) is 0
        # if len(display_settings["list_of_mask"])>1:
        #     for idx, mask in enumerate(display_settings["list_of_mask"]):
        #         if np.max(mask[...,:3]) == 0:
        #             display_settings["list_of_mask"].pop(idx)

    def captureManager(self):
        self.imageFrozen = not self.imageFrozen
        if self.imageFrozen:
            self.mainUI.btn_capture.setText("Release \n(Spacebar)")
            self.deviceManager.image_captured_paused = True
        else:
            self.mainUI.btn_capture.setText("Capture \n(Spacebar)")
            self.deviceManager.image_captured_paused = False

    def updateSliderBrushStrength(self):
        """Handle the slider value change and update the brush size."""
        # Get the new value of the slider
        brush_settings["opacity"] = self.mainUI.brush_slider.value()

        # Optionally, update the UI or take other actions
        # print(f"New brush size: {brush_settings["size"]}")
        self.update_image_display()  # Update the image to reflect the new brush size

    def updateSliderBinaryThreshold(self):
        """Handle the slider value change and update the binary mask threshold."""
        # Get the new value of the slider
        _thres = self.mainUI.binarythres_slider.value()
        # remap the value from 0-99 into an integer between 0-255
        brush_settings["binary_mask_thres"] = int(_thres * 2.55)
        if self.binary_mask is not None:
            self.binary_mask.threshold = brush_settings["binary_mask_thres"]
            

        # Optionally, update the UI or take other actions

    def update_webcam_feed(self):
        """Capture and display a new frame from the webcam."""
        if os_settings["webcam_mode"]:
            if not self.imageFrozen:
                display_settings["image"] = self.deviceManager.get_image_from_webcam()  # Load first frame from webcam
                # display_settings["image"] = cv2.resize(display_settings["image"], (W, H))  # Resize image to fit the window
        else:
            display_settings["image"] = self.editorManager.get_image_from_folder()  # Load first frame from webcam
        self.update_image_display()

    
    def update_image_display(self):  
        """Overlay the mask on top of the image using the mask's values wherever they are non-zero."""
        # Create a copy of the original image
        rendered_image = display_settings["image"].copy()

        # create a merged copy of all the masks to avoid modifying the original, allowing felixibility in visualizations
        rendered_mask = merge_masks(display_settings["list_of_mask"])
        
        # MASK_PIXELS_FILTER = Identify the color pixels that have some color in the RENDERED_MASK
        # Blend the mask's RGB channels with the original image
        # handle brush preview 
        
        # MASK_PIXELS_FILTER = Identify the color pixels that have some color in the RENDERED_MASK
        mask_pixels_filter = rendered_mask[...,:3] 
        mask_pixels_filter = np.any(mask_pixels_filter != 0, axis = -1)
        
        new_brush_intensity = brush_settings["opacity"]/100
        # Blend the mask with the image
        # Blend the mask's RGB channels with the original image
        rendered_image[mask_pixels_filter, :3] = (
            new_brush_intensity * rendered_mask[mask_pixels_filter, :3] +  # Mask color contribution
            (1 - new_brush_intensity) * rendered_image[mask_pixels_filter, :3]  # Original image contribution
        ).astype(np.uint8)  # Ensure the result stays within valid RGB values

        # BRUSH MODE
        if brush_settings["is_brush_mode"] == "brush":
            # Handle brush preview (if not drawing)
            if not self.drawing:
                preview_image = rendered_image.copy()
                if not self.current_pos.isNull():
                    # Draw the brush preview circle
                    cv2.circle(preview_image, (self.current_pos.x(), self.current_pos.y()), brush_settings["size"] // 2,
                            self.brush_preview_color, 2)
                qimage = self.convert_cv_qt(preview_image)
            else:
                qimage = self.convert_cv_qt(rendered_image)

            # Set the QPixmap on the QLabel
            self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
            
        # POLYGON MODE 
        elif brush_settings["is_brush_mode"] == "polygon":
            polygon = self.polygon_manager.current_polygon
            preview_image = rendered_image.copy()
            for idx,point in enumerate(polygon["POINTS"]):
                if idx < len(polygon["POINTS"])-1:
                    cv2.line(preview_image, point, polygon["POINTS"][idx+1], color=brush_settings["color"], thickness=brush_settings["thickness"])
                cv2.circle(preview_image, (point.x, point.y), 2, brush_settings["color"], -1)
                        

            qimage = self.convert_cv_qt(preview_image)

            # Set the QPixmap on the QLabel
            self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
        
        
        # FREEHAND MODE
        elif brush_settings["is_brush_mode"] == "freehand":
            polygon = self.freehand_manager.current_polygon
            preview_image = rendered_image.copy()
            for idx,point in enumerate(polygon["POINTS"]):
                if idx < len(polygon["POINTS"])-1:
                    cv2.line(preview_image, point, polygon["POINTS"][idx+1], color=brush_settings["color"], thickness=brush_settings["thickness"])
                cv2.circle(preview_image, (point.x, point.y), 2, brush_settings["color"], -1)
                        

            qimage = self.convert_cv_qt(preview_image)

            # Set the QPixmap on the QLabel
            self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
        
        # BEZIER mode
        elif brush_settings["is_brush_mode"] == "bezier":
            preview_image = rendered_image.copy()
            if not self.bezier_manager is None and len(self.bezier_manager.lobu) >0:
                self.bezier_manager.draw(self.bezier_current_pos,preview_image)       

                qimage = self.convert_cv_qt(preview_image)

                # Set the QPixmap on the QLabel
                self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
            else:
                qimage = self.convert_cv_qt(preview_image)
                self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
        
        elif brush_settings["is_brush_mode"] == "bucket":
            qimage = self.convert_cv_qt(rendered_image)
            self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
        elif brush_settings["is_brush_mode"] == "binmask":
            # pass the preview image to the binary mask manager
            preview_image = self.binary_mask.display(rendered_image.copy())
            qimage = self.convert_cv_qt(preview_image)
            self.mainUI.imageDisplay.setPixmap(QPixmap.fromImage(qimage))
            
    def convert_cv_qt(self, cv_img):
        """Convert from an OpenCV image to QImage."""
        rgb_image = cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB)  # Convert BGR to RGB
        h, w, ch = rgb_image.shape
        bytes_per_line = ch * w
        return QImage(rgb_image.data, w, h, bytes_per_line, QImage.Format_RGB888)

    def map_to_image_display(self, global_pos):
        """Convert global mouse position to position relative to image display, considering potential scaling."""
        # Map global position to the image display's local coordinates
        local_pos = self.mainUI.imageDisplay.mapFromGlobal(global_pos)
        label_width = self.mainUI.imageDisplay.width()
        label_height = self.mainUI.imageDisplay.height()
        image_height, image_width = display_settings["image"].shape[:2]
        
        # Scale mouse position based on how the image is resized
        tx = local_pos.x()/label_width
        ty = local_pos.y()/label_height
        mapped_x = int(tx * label_width)
        mapped_y = int(ty * label_height)


        # Ensure the position is within the image bounds
        mapped_x = max(0, min(mapped_x, image_width - 1))
        mapped_y = max(0, min(mapped_y, image_height - 1))


        return QPoint(mapped_x, mapped_y)

    def mousePressEvent(self, event):
        """Start drawing when the mouse is pressed."""
        if cursor_settings["in_display"]:
            if brush_settings["is_brush_mode"] == "brush":
                # ################## <BRUSH MOUSE PRESS ENTER EVENT> #######################
                
                if event.button() == Qt.LeftButton:
                    global_pos = event.globalPos()  # Get global position
                    self.last_point = self.map_to_image_display(global_pos)
                    if not os_settings["substractive_mode"]:
                        cv2.circle(display_settings["list_of_mask"][-1], (self.last_point.x(), self.current_pos.y()), brush_settings["size"] // 2,
                                    brush_settings["color"],-1)
                    else:
                        cv2.circle(display_settings["list_of_mask"][-1], (self.last_point.x(), self.current_pos.y()), brush_settings["size"] // 2,
                                    (0,0,0,0),-1)
                # ################## </BRUSH MOUSE PRESS ENTER EVENT> #######################
            # if polygon mode
            elif brush_settings["is_brush_mode"] == "polygon":
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                local_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.polygon_manager.current_polygon["POINTS"].append(local_pos)
                
            elif brush_settings["is_brush_mode"] == "freehand":
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                local_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.freehand_manager.startFreehand(local_pos)
                
            elif brush_settings["is_brush_mode"] == "bezier":
                self.bezier_manager.onMouseEventDown(self.bezier_current_pos)
                
            elif brush_settings["is_brush_mode"] == "bucket":
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                local_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.bucket_manager.onMouseEventDown(local_pos)
            
            elif brush_settings["is_brush_mode"] == "binmask":
                self.binary_mask.onMouseEventDown()
                
    def mouseMoveEvent(self, event):
        """Track mouse movement and update the brush preview."""
        if cursor_settings["in_display"]:
            if brush_settings["is_brush_mode"] == "brush":
                global_pos = event.globalPos()  # Get global position of the mouse
                self.current_pos = self.map_to_image_display(global_pos)
                ############# TODO: ENCAPSULATE BRUSH INTO OBJECT ###################
                if event.buttons() & Qt.LeftButton and not os_settings["substractive_mode"]:
                    # Update the mask with the brush stroke in OpenCV
                    cv2.line(display_settings["list_of_mask"][-1], (self.last_point.x(), self.last_point.y()), 
                            (self.current_pos.x(), self.current_pos.y()), brush_settings["color"], brush_settings["size"])
                    self.last_point = self.current_pos
                    # display_settings["list_of_mask"].append(np.zeros_like(display_settings["list_of_mask"][-1]))

                elif event.buttons() & Qt.LeftButton and os_settings["substractive_mode"]:
                    # Erase by drawing over the mask with a transparent color
                    cv2.line(display_settings["list_of_mask"][-1], (self.last_point.x(), self.last_point.y()), 
                            (self.current_pos.x(), self.current_pos.y()), (255,255,255,255), brush_settings["size"])
                    self.last_point = self.current_pos
                    # display_settings["list_of_mask"].append(np.zeros_like(display_settings["list_of_mask"][-1]))
                ############# END: ENCAPSULATE BRUSH INTO OBJECT ###################
                
            elif cursor_settings["in_display"] and brush_settings["is_brush_mode"] == "bezier":
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                self.bezier_current_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.bezier_manager.update(self.bezier_current_pos)
            
            elif cursor_settings["in_display"] and brush_settings["is_brush_mode"] == "freehand":
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                self.freehand_current_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.freehand_manager.updatePolyOnMove(self.freehand_current_pos)
            
            self.update_image_display()
            
        
        

    def mouseReleaseEvent(self, event):
        """Finish drawing when the mouse is released."""
        if cursor_settings["in_display"]:
            if brush_settings["is_brush_mode"] == "brush":                
                if event.button() == Qt.LeftButton:
                    self.drawing = False
                    ############# TODO: ENCAPSULATE BRUSH INTO OBJECT ###################
                    if os_settings["substractive_mode"]:
                        # create a new mask from the selected blob
                        _boolean_mask = np.copy(display_settings["list_of_mask"][-1])
                        # invert the boolean mask so we keep everything EXCEPT the polygon
                        inverted_mask = cv2.bitwise_not(_boolean_mask)      
                        
                        if os_settings["top_layer_edit"]:
                            # Modify the mask on the top layer only (second to last because we added a new layer)
                            display_settings["list_of_mask"][-2] = cv2.bitwise_and(display_settings["list_of_mask"][-2], inverted_mask)
                            display_settings["list_of_mask"].pop(-1)  # Remove the last mask
                            _ = add_empty_mask()  # Add a new empty mask to the list
                            
                        else:
                            # iterate over all masks erasing the pixels delimited by this mask
                            for idx, mask in enumerate(display_settings["list_of_mask"]):
                                display_settings["list_of_mask"][idx] = cv2.bitwise_and(mask, inverted_mask)
                        
                    ############# END: ENCAPSULATE BRUSH INTO OBJECT ###################

            if brush_settings["is_brush_mode"] == "bezier":
                self.bezier_manager.onMouseEventUp(self.bezier_current_pos)
                global_pos = event.globalPos()  # Get global position
                self.last_point = self.map_to_image_display(global_pos)
                self.bezier_current_pos = Coordinate(self.last_point.x(), self.last_point.y())
                self.bezier_manager.update(self.bezier_current_pos)
            if brush_settings["is_brush_mode"] == "freehand":
                if event.button() == Qt.LeftButton:
                    self.freehand_manager.finishPolygon()
                    
    def wheelEvent(self, event):
        """Handle mouse scroll events."""
        if cursor_settings["in_display"] and brush_settings["is_brush_mode"]=="brush":
            delta = event.angleDelta().y()  # Get the vertical scroll amount
            if delta > 0:
                brush_settings["size"] += brush_settings["resize_sensitivity"]
            elif delta < 0:
                brush_settings["size"] -= 3
                brush_settings["size"] = 1 if brush_settings["size"] <= 1 else brush_settings["size"]

            self.update_image_display()  # Update the display to reflect the new brush size
        
    def keyReleaseEvent(self, event):
        if brush_settings["is_brush_mode"] == "polygon":
            # Check if the Enter key was pressed
            if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
                self.polygon_manager.finishPolygon()
                
                
        elif brush_settings["is_brush_mode"] == "bezier":
            if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
                self.bezier_manager.finishBezier()
            
            # else:
            #     super().keyPressEvent(event)  # Pass the event to the parent class for default behavior
        elif brush_settings["is_brush_mode"] == "binmask":
            if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
                self.binary_mask.finish_binary_mask()

    def closeEvent(self, event):
        # This function is triggered when the window is about to close
        reply = QMessageBox.question(self, 'Window Close', 
                                    'Would you like to save the project before closing the window?',
                                    QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel, 
                                    QMessageBox.Cancel)

        if reply == QMessageBox.Yes:
            # Save the project and close the window
            self.mainUI.save_json_file()
            self.videoDeviceManager.release_video_device()
            event.accept()  # Accept the event to close the window
        elif reply == QMessageBox.No:
            # Do not save, just close the window
            event.accept()  # Accept the event to close the window
        else:
            # Cancel the close action
            event.ignore()  # Ignore the close event, keeping the window open


# wrapping everything into a main function to allow for easier integration via PyPI and CLI
def main(argv=None):
    """Entry point for CLI and `python -m livepyxel`."""
    argv = argv or sys.argv

    # High-DPI settings (must be set before QApplication is created)
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

    app = QApplication(argv)

    # Construct and show your main window
    window = ImageAnnotator()
    window.show()

    return app.exec_()
        

if __name__ == "__main__":
    raise SystemExit(main())


import math
import numpy as np
from .constants import *

SUBDIVS = 150 #total number of divisions in a bezier curve
SEGMENT_DISTANCE = 1/SUBDIVS #how long is a segment relative to a unit

# DD. POINT
# pt = Coordinate(int,int)
# interp. a point in the scaled view of a webcam feed, scaled to fit the imageDisplay Widget of the main program
pt = Coordinate(0,0)

# DD. BEZIER
# bezier = Bu()
# interp. an object representing a Bezier unit that contains:
# - point 1
# - point 2 (anchor)
# - point 3

# At the moment of its conception, the Bu will start at a point A
class Bu():
    def __init__(self,pA=(0,0),pG=(0,0),pB=(0,0), keep_pG=False):
        self.keep_pG = keep_pG
        self.pA = pA
        self.pG = pA
        self.pG_inverse = self.pG
        self.pB = self.pA #set the starting and end position of the line at the same point at the beginning
        self.ptA_Set = True
        self.ptB_Set = False #determines whether the points G and B that make the unit should change
        self.finishedBu = False
        self.draw_laterals = False #activates when the lateral lines reflecting gravity points have to show up
        self.saved_points = []
    
    def fill_bezier_unit(self,pA,pG,pB):
        saved_points = []
        startingX = pA.x
        startingY = pA.y
        for i in range(SUBDIVS+1):
            saved_points.append((Coordinate(int(startingX),int(startingY))))
            
            # Calculate the relative distance traveled in the time i starting at A
            x1 = pA.x + (pG.x - pA.x) * (i * SEGMENT_DISTANCE)
            y1 = pA.y + (pG.y - pA.y) * (i * SEGMENT_DISTANCE)

            x2 = pG.x + (pB.x - pG.x) * (i * SEGMENT_DISTANCE)
            y2 = pG.y + (pB.y - pG.y) * (i * SEGMENT_DISTANCE)

            x = x1 + (x2 - x1) * (i * SEGMENT_DISTANCE)
            y = y1 + (y2 - y1) * (i * SEGMENT_DISTANCE)

            # pygame.draw.line(display,brush_settings["color"],(startingX,startingY),(x,y),3)

            startingX = x
            startingY = y
        # Add the last line to the bezier spline
        saved_points.append((Coordinate(int(startingX),int(startingY))))
        return saved_points

        
    
    def draw(self, coor, preview_image):
        # as long as user drags without pressing down mouse for second time, the point G will be the same as the point A (i.e. no deformation)
        if self.ptA_Set and not self.ptB_Set:
            self.pB = coor 
            if not self.keep_pG:
                self.pG = coor
        # RENDER THE LINE SEGMENTS OF THE BEZIER CURVE
        self.saved_points = self.fill_bezier_unit(self.pA,self.pG,self.pB)

        
        for idx,point in enumerate(self.saved_points[:-1]):
            cv2.line(preview_image, point, self.saved_points[idx+1], color=brush_settings["color"], thickness=brush_settings["thickness"])
            # cv2.circle(preview_image, (point.x, point.y), 2, brush_settings["color"], -1)
        
        if self.draw_laterals:
            # calculate the distance between pB and the cursor
            mx, my = coor
            a = mx - self.pB.x
            o = my - self.pB.y
            radius = (a**2 + o**2)**0.5
            # If the radius over 5, let's assume the user wants to update pG, in which case we just calculate the inverse pG_inverse
            if radius > 2:
                # get the angle between the point B and the position of the cursor
                angle = math.atan2(o,a)
                # calculate the position in x,y for the pointG and pointG_inverse, using the distance pB-cursor as radius
                x_G_inverse = self.pB[0] + (math.cos(angle) * radius)
                y_G_inverse = self.pB[1] + (math.sin(angle) * radius)
                x_G = self.pB[0] - (math.cos(angle) * radius)
                y_G = self.pB[1] - (math.sin(angle) * radius)
                self.pG = Coordinate(int(x_G),int(y_G))
                self.pG_inverse = Coordinate(int(x_G_inverse),int(y_G_inverse))
                
                
                

            else:
                # get the angle between the point B and the position of the pointG
                static_a = self.pB[0] - self.pG[0]
                static_o = self.pB[1] - self.pG[1]
                dist_B_G = (static_a**2 + static_o**2) ** 0.5
                angle_BG = math.atan2(static_o,static_a)
                x_G_inverse = self.pB[0] + (math.cos(angle_BG) * dist_B_G)
                y_G_inverse = self.pB[1] + (math.sin(angle_BG) * dist_B_G)
                # self.pG = Coordinate(self.pG[0],self.pG[1])
                self.pG_inverse = Coordinate(int(x_G_inverse),int(y_G_inverse))

            cv2.line(preview_image, self.pB, self.pG, color=brush_settings["color"], thickness=brush_settings["thickness"])
            cv2.circle(preview_image, self.pG, 2, brush_settings["color"], -1)
            
            cv2.line(preview_image, self.pB, self.pG_inverse, color=brush_settings["color"], thickness=brush_settings["thickness"])
            cv2.circle(preview_image, self.pG_inverse, 2, brush_settings["color"], -1)

    # FD. savePoint()
    # Signature: None -> None
    # purp. calculate the position of every item in that makes the Bezier unit and save the information into a list for drawing
    def savePoints(self):
        self.saved_points = []
        startingX = self.pA[0]
        startingY = self.pA[1]
        # self.saved_points.append((startingX,startingY))
        for i in range(SUBDIVS):
            # Calculate the relative distance traveled in the time i starting at A
            x1 = self.pA[0] + (self.pG[0] - self.pA[0]) * (i * SEGMENT_DISTANCE)
            y1 = self.pA[1] + (self.pG[1] - self.pA[1]) * (i * SEGMENT_DISTANCE)

            x2 = self.pG[0] + (self.pB[0] - self.pG[0]) * (i * SEGMENT_DISTANCE)
            y2 = self.pG[1] + (self.pB[1] - self.pG[1]) * (i * SEGMENT_DISTANCE)

            x = x1 + (x2 - x1) * (i * SEGMENT_DISTANCE)
            y = y1 + (y2 - y1) * (i * SEGMENT_DISTANCE)

            # pygame.draw.line(display,(random.randint(0,255),random.randint(0,255),random.randint(0,255)),(startingX,startingY),(x,y),3)
            
            startingX = x
            startingY = y
            self.saved_points.append((startingX,startingY))
            
                
 
# DD. SPLINE
# spline = Spline()
# interp. the collection of Bezier units that create a spline
class Bezierman():
    def __init__(self):
        self.first_click = True
        self.lobu = []
        self.doneSpline = False
    
    def draw(self, coor, preview_image):
        ########## draw each bezier unit
        for bu in self.lobu:
            bu.draw(coor, preview_image)
        
    def onMouseEventUp(self, coor):
        # if self.first_click:
        #     self.first_click = False

        if len(self.lobu)>0:
            self.lobu[-1].finishedBu = True #THE LAST ELEMENT IS THE PREVIOUS BEZIER UNIT, that has already been finished
        else:
            bu = Bu(coor)
            self.lobu.append(bu)
                
    def onMouseEventDown(self, coor):
        if len(self.lobu)>0:
            last_bu = self.lobu[-1] #active bezier unit is last in the list
            if last_bu.ptA_Set and not last_bu.ptB_Set and not last_bu.finishedBu:
                last_bu.ptB_Set = True
        
    
    def update(self, coor):
        if len(self.lobu)>0:
            last_bu = self.lobu[-1] #the active bezier unit
            if last_bu.ptA_Set and not last_bu.ptB_Set and not last_bu.finishedBu:
                last_bu.pB = coor
            elif last_bu.ptA_Set and last_bu.ptB_Set and not last_bu.finishedBu:
                # Draw gravity lines using pointer and draw Bezier curves
                last_bu.draw_laterals = True
                # last_bu.pB = pygame.mouse.get_pos()
            elif last_bu.ptA_Set and last_bu.ptB_Set and last_bu.finishedBu:
                # if not self.doneSpline:
                last_bu.draw_laterals = False
                if len(self.lobu)%2 != 0:
                    bu = Bu(last_bu.pB, keep_pG=True)
                else:
                    bu = Bu(last_bu.pB, keep_pG=False)
                    
                # if there's already other Bu's, use the previous (i = -1) Bu
                # variable lastbu.pG for this pG pos
                bu.pG = last_bu.pG_inverse
                self.lobu.append(bu)
    
    def finishBezier(self):
        if brush_settings["is_brush_mode"] == "bezier":
            final_points = []
            for bu in self.lobu[:-1]:
                final_points += bu.saved_points                
            # points = [(pt.x, pt.y) for pt in self.current_polygon["POINTS"]]
            points = np.array([[p.x, p.y] for p in final_points], np.int32)
            points = points.reshape((-1, 1, 2))
            # evaluate if there are any points to annotate
            if len(points) == 0:
                return

            if os_settings["substractive_mode"]:
                # create a new mask from the selected polygon
                _boolean_mask = np.zeros_like(display_settings["list_of_mask"][-1])
                # fill it with the color (255,255,255) to erase the pixels at a later stage
                cv2.fillPoly(_boolean_mask, [points], color=(255,255,255))
                # invert the boolean mask so we keep everything EXCEPT the polygon
                inverted_mask = cv2.bitwise_not(_boolean_mask)
                # iterate over all masks erasing the pixels delimited by this mask, depending upon user's selection
                if os_settings["top_layer_edit"]:
                    display_settings["list_of_mask"][-1] = cv2.bitwise_and(display_settings["list_of_mask"][-1], inverted_mask)
                else:
                    for idx, mask in enumerate(display_settings["list_of_mask"]):
                        display_settings["list_of_mask"][idx] = cv2.bitwise_and(mask, inverted_mask)
            else:
                
                
                _mask = add_empty_mask()
                # fill the color with the brush color selected by the user and save it, then create a new mask
                cv2.fillPoly(_mask, [points], color=brush_settings["color"])
                # display_settings["list_of_mask"].append(np.zeros_like(display_settings["list_of_mask"][-1]))
                
            # self.current_polygon = {"DONE":False, "POINTS":[], "COLOR":None}
            self.lobu = []
            
    def pop_last_point(self):
        if len(self.lobu)>1:
            self.lobu = self.lobu[:-2]
   


import numpy as np
from .constants import *

# DD. POINT
# pt = Coordinate(int,int)
# interp. a point in the scaled view of a webcam feed, scaled to fit the imageDisplay Widget of the main program
pt = Coordinate(0,0)

# DD. BUCKET_MANAGER
# bucketman = Bucketman()
# interp. a set of attributes used to define the behavior of the bucket tool
class Bucketman():
    def __init__(self, parent=None):
        # Store a reference to the parent ImageAnnotator
        self.parent = parent
    
    def onMouseEventDown(self, coor):
        """
        Implements a bucket/fill tool that changes the color of an entire annotation.
        
        Parameters:
        coor (QPoint): The coordinates where the user clicked
        """
        # Convert QPoint to a tuple of (x, y)
        x, y = coor.x, coor.y
        
        # Find which mask contains the annotation at the clicked point
        target_mask_index = None
        
        for idx, mask in enumerate(display_settings["list_of_mask"]):
            # Check if there's color at the clicked position in this mask
            if np.any(mask[y, x] != 0):
                target_mask_index = idx
                break
        
        # If no colored pixel was found at the clicked position, return
        if target_mask_index is None:
            return
        
        # Get the mask that contains the clicked annotation
        target_mask = display_settings["list_of_mask"][target_mask_index]
        
        # Create a binary mask where any non-zero pixel will be changed
        # This identifies all pixels that are part of the annotation
        binary_mask = np.any(target_mask != 0, axis=2)
        
        # Apply the new color to all non-zero pixels in the mask
        # If using subtractive mode, set to black (erase)
        if os_settings["substractive_mode"]:
            # In subtractive mode, we set the color to black (effectively removing it)
            target_mask[binary_mask] = (0, 0, 0, 0)
        else:
            # In normal mode, we set the color to the currently selected color
            # r, g, b = brush_settings["color"]
            target_mask[binary_mask] = brush_settings["color"]
        
        # Update the mask in the list
        display_settings["list_of_mask"][target_mask_index] = target_mask
        
        # # Update the display to show the changes
        # if self.parent:
        #     self.parent.update_image_display()


# import pygame 
import os 
import cv2 
import argparse
from os.path import join as jn
from collections import namedtuple

import random
import json 

import sys
import cv2
import numpy as np
from PyQt5.QtCore import Qt, QPoint, QTimer
from PyQt5.QtGui import QImage, QPixmap, QKeySequence
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QApplication, QLabel, QMainWindow, QPushButton,QColorDialog,QMessageBox,QShortcut
import uuid
########################### ARGPARSE arguments ###################################
parser = argparse.ArgumentParser(description='Line measurement tool')
parser.add_argument("--ci", default=0,type=int,help='Index of the camera (for multiple camera devices connected)')
parser.add_argument("--res", default=0.8,type=float,help="Determine the resolution of the camera from 0.1 to 1.0")
args = parser.parse_args()

##################################################################################
# pygame.init()
# display_info = pygame.display.Info()
# define the dimensions of the screen
# W = int(display_info.current_w*args.res)
# H = int(W*cameraRatio)
# W = 1920
# H = 1080



# Determine the width and height ratio of the camera
cap = cv2.VideoCapture(args.ci)
cameraW = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
cameraH = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
cameraRatio = cameraH/cameraW #what proportion of the width is the height (e.g. for 1920 x 1080, it's 0.56)


# Define the dimensions of the screen
W = 1920
H = 1080


# DD LATERAL_PADDING_MENU
# LATERAL_PADDING = (int,int)
# interp. how many pixels to leave as padding to accomodate the lateral panel of the application
# Define the named tuple with fields 'x' and 'y'
Coordinate = namedtuple('Coordinate', ['x', 'y'])
LATERAL_PADDING = Coordinate(400,0)
TOP_PADDING = Coordinate(0,80)


# DD. BRUSH_MAX_MEMORY
# brush_max_memory = int
# interp. the maximum number of brush strokes that can be stored in memory
BRUSH_MAX_MEM = 3

# DD. SAVEPATH 
# savePath = str
# interp. the location of the file that will store the program's output
savePath = jn(os.path.dirname(__file__),"output.txt")

# DD. BRUSH_SETTINGS
# brush_settings = {"color":(int, int, int)}
# interp. a set of parameters represented as a mutable object to be used globally within different submodules of the program
brush_settings = {"color":(0, 255, 0), "size":20,"resize_sensitivity":5, "is_brush_mode":"brush", "brush_strokes_in_memory_counter":0,"opacity":0,"color_before_substracting_mode":None,"thickness":1, "binary_mask_thres":0}

# DD. DISPLAY_SETTINGS
# display_settings = {"image":np.array, "mask":np.array, "statusBar":QtWidgets.QStatusBar}
# interp. a set of numpy arrays representing images and masks for the program
display_settings = {"image":None, "mask":None,"list_of_mask":None, "statusBar":None}

cursor_settings = {"in_display":False}

# substractive mode turns the brush or tool into an eraser
os_settings = {"directory":"", "config":"", "masks_path":"","images_path":"","substractive_mode":False, "top_layer_edit":False, "webcam_mode":True}



# FD. merge_masks()
# purp. merge multiple image arrays where higher index images take precedence over lower index ones,
# but only if their pixel values are non-zero.
def merge_masks(list_of_masks):
        """
        Created by Claude Sonnet v.3.2 on March 6th, 2025
        Merge multiple image arrays where higher index images take precedence over lower index ones,
        but only if their pixel values are non-zero.
        
        Assumes:
        - All images in display_settings["list_of_mask"] are the same size
        - At least one image is always provided
        - The image list is stored in display_settings["list_of_mask"]
        
        Returns:
        rendered_mask: NumPy array containing the merged result
        """
        # Initialize the rendered mask with the first image
        rendered_mask = np.copy(display_settings["list_of_mask"][0])
        
       
        # Merge the images according to the specified rules
        for img in display_settings["list_of_mask"][1:]:
            # Create a mask of non-zero pixels in the current image
            non_zero_mask = np.any(img != 0, axis=-1)
            
            # Update the rendered mask with non-zero pixels from the current image
            rendered_mask[non_zero_mask] = img[non_zero_mask]
        
        return rendered_mask


# def show_status_message(message, is_success=False, is_error=False, timeout=5000):
#     """Displays a styled status message with optional icon on the status bar."""
#     statusBar = display_settings["statusBar"]
#     statusBar().showMessage(message, timeout)  # Shows for 5 seconds
                
#     # Optional: Flash the status bar for better visibility
#     palette = statusBar().palette()
#     original_color = palette.color(palette.Window)
#     palette.setColor(palette.Window, QtGui.QColor(0, 150, 0))  # Green background
#     statusBar().setPalette(palette)
    
    

def show_status_message(message, is_success=False, is_error=False, timeout=5000):
    """Displays a styled status message with optional icon on the status bar."""
    status_bar = display_settings["statusBar"]
    
    # Clear existing message
    status_bar().clearMessage()

    # Compose text with icon
    if is_success:
        icon_path = "icons/success_icon.png"
    elif is_error:
        icon_path = "icons/fail_icon.png"
    else:
        icon_path = None

    if icon_path and os.path.exists(icon_path):
        pixmap = QtGui.QPixmap(icon_path).scaled(16, 16, QtCore.Qt.KeepAspectRatio)
        icon = QtWidgets.QLabel()
        icon.setPixmap(pixmap)
        message_label = QtWidgets.QLabel(message)
        
        container = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(container)
        layout.setContentsMargins(5, 0, 5, 0)
        layout.addWidget(icon)
        layout.addWidget(message_label)
        status_bar().addPermanentWidget(container)

        # Remove widget after timeout
        QtCore.QTimer.singleShot(timeout, lambda: (
            status_bar().removeWidget(container),
            container.deleteLater()
        ))
    else:
        # Just show plain message
        status_bar().showMessage(message, timeout)

    # Change background color
    palette = status_bar().palette()
    if is_success:
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor(220, 255, 220))  # light green
    elif is_error:
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor(255, 220, 220))  # light red
    else:
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor(240, 240, 240))  # neutral
    status_bar().setAutoFillBackground(True)
    status_bar().setPalette(palette)

    # Reset color after timeout
    QtCore.QTimer.singleShot(timeout, lambda: (
        status_bar().setPalette(QtWidgets.QApplication.palette()),
        status_bar().setAutoFillBackground(False)
    ))
    
def add_empty_mask():
    display_settings["list_of_mask"].append(np.zeros_like(display_settings["list_of_mask"][-1]))
    return display_settings["list_of_mask"][-1]


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_V4.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from os.path import join as jn
import os
from .constants import *
from .edit_label_box import Ui_Dialog
from .warning_dialog_delete import Ui_Dialog as Ui_Dialog_Delete

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("LivePyxel")
        MainWindow.resize(1536, 672)
        MainWindow.setMinimumSize(QtCore.QSize(1536, 672))
        MainWindow.setStyleSheet("background-color: rgb(30, 30, 30);\n"
"color: rgb(255, 255, 255);")
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        # Webcam feed
        self.script_path = os.path.dirname(os.path.abspath(__file__))
        print(self.script_path)
        [print(x) for x in os.listdir(self.script_path)]
        self.imageDisplay = WebcamWidget(self.centralwidget)
        self.imageDisplay.setMinimumSize(QtCore.QSize(720, 480))
        self.imageDisplay.setMaximumSize(QtCore.QSize(720, 480))
        self.imageDisplay.setStyleSheet("background-color: rgb(125,125,125);")
        self.imageDisplay.setText("")
        self.imageDisplay.setObjectName("imageDisplay")
        self.gridLayout_2.addWidget(self.imageDisplay, 2, 2, 1, 1)
        # Scrolling area
        self.scrollArea = QtWidgets.QScrollArea(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setMaximumSize(QtCore.QSize(300, 600))
        self.scrollArea.setStyleSheet("background-color: rgb(60, 60, 60);\n""color: rgb(255, 255, 255);\n""")
        self.scrollArea.setFrameShadow(QtWidgets.QFrame.Raised)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents_3 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_3.setGeometry(QtCore.QRect(0, 0, 298, 478))
        self.scrollAreaWidgetContents_3.setObjectName("scrollAreaWidgetContents_3")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents_3)
        self.gridLayout_6.setVerticalSpacing(20)
        self.gridLayout_6.setObjectName("gridLayout_6")
        # Create a new button
        self.label_block_0 = QtWidgets.QWidget(self.scrollAreaWidgetContents_3) #place a new widget as child of the scroll area
        # define new button block's policy and attributes
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_block_0.sizePolicy().hasHeightForWidth())
        self.label_block_0.setSizePolicy(sizePolicy)
        self.label_block_0.setMaximumSize(QtCore.QSize(16777215, 40))
        self.label_block_0.setObjectName("label_block_0")
        # button block has a layout as well
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout(self.label_block_0)
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        # add a label to the button block's layout
        self.label_select_0 = QtWidgets.QLabel(self.label_block_0)
        self.label_select_0.setObjectName("label_select_0")
        self.horizontalLayout_13.addWidget(self.label_select_0)
        # add a select button to the button block's layout
        self.btn_select_0 = QtWidgets.QPushButton(self.label_block_0)
        self.btn_select_0.setMaximumSize(QtCore.QSize(60, 16777215))
        self.btn_select_0.setStyleSheet("background-color: rgb(0, 255, 0);")
        self.btn_select_0.setText("")
        self.btn_select_0.setObjectName("btn_select_0")
        self.horizontalLayout_13.addWidget(self.btn_select_0)
        self.btn_select_0.clicked.connect(lambda: self.traverse_select_buttons(self.btn_select_0))
        # add an edit button to the button block's layout
        self.btn_edit_0 = QtWidgets.QPushButton(self.label_block_0)
        self.btn_edit_0.setMinimumSize(QtCore.QSize(20, 0))
        self.btn_edit_0.setMaximumSize(QtCore.QSize(30, 16777215))
        self.btn_edit_0.setText("")
        self.icon_edit = QtGui.QIcon()
        self.icon_edit.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/edit_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_edit_0.setIcon(self.icon_edit)
        self.btn_edit_0.setObjectName("btn_edit_0")
        self.horizontalLayout_13.addWidget(self.btn_edit_0)
        self.btn_edit_0.clicked.connect(lambda: self.traverse_edit_buttons(self.btn_edit_0))
        # add a remove button to the button block's layout
        self.btn_remove_0 = QtWidgets.QPushButton(self.label_block_0)
        self.btn_remove_0.setMinimumSize(QtCore.QSize(20, 0))
        self.btn_remove_0.setMaximumSize(QtCore.QSize(30, 16777215))
        self.btn_remove_0.setText("")
        self.icon_trash = QtGui.QIcon()
        self.icon_trash.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/trash_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btn_remove_0.setIcon(self.icon_trash)
        self.btn_remove_0.setObjectName("btn_remove_0")
        self.horizontalLayout_13.addWidget(self.btn_remove_0)
        self.btn_remove_0.setDisabled(True)
        self.btn_remove_0.setStyleSheet("background-color:rgb(45,45,45)")
        # self.btn_remove_0.clicked.connect(lambda: self.traverse_remove_buttons(self.btn_remove_0))
        
        
        
        self.gridLayout_6.addWidget(self.label_block_0, 1, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents_3)
        self.gridLayout_2.addWidget(self.scrollArea, 2, 0, 1, 1)
        
        # Brush slider Settings — using manual geometry
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.frame.setMinimumSize(QtCore.QSize(300, 80))  # same as scroll panel
        self.frame.setMaximumSize(QtCore.QSize(300, 80))  # lock width to match scroll area
        self.frame.setStyleSheet("background-color: transparent;")

        # First label and slider (e.g., Mask Opacity)
        self.label_6 = QtWidgets.QLabel(self.frame)
        self.label_6.setGeometry(QtCore.QRect(10, 0, 200, 16))  # wider label
        self.label_6.setText("Mask opacity")
        self.label_6.setObjectName("label_6")

        self.brush_slider = QtWidgets.QSlider(self.frame)
        self.brush_slider.setGeometry(QtCore.QRect(10, 20, 280, 20))
        self.brush_slider.setOrientation(QtCore.Qt.Horizontal)
        self.brush_slider.setObjectName("brush_slider")

        # Second label and slider (e.g., Binary Mask Threshold)
        self.label_7 = QtWidgets.QLabel(self.frame)
        self.label_7.setGeometry(QtCore.QRect(10, 40, 200, 16))
        self.label_7.setText("binary mask threshold")
        self.label_7.setObjectName("label_7")

        self.binarythres_slider = QtWidgets.QSlider(self.frame)
        self.binarythres_slider.setGeometry(QtCore.QRect(10, 60, 280, 20))
        self.binarythres_slider.setOrientation(QtCore.Qt.Horizontal)
        self.binarythres_slider.setObjectName("binarythres_slider")

        # Add to layout
        self.gridLayout_2.addWidget(self.frame, 0, 0, 1, 1)
        
        
        # Add section
        self.btn_add = QtWidgets.QPushButton(self.centralwidget)
        self.btn_add.setStyleSheet("background-color: rgb(80, 120, 80);")
        self.btn_add.setObjectName("btn_add")
        self.gridLayout_2.addWidget(self.btn_add, 3, 0, 1, 1)
        
        
        
        # Tip right panel with webcam related buttons
        self.widget_2 = QtWidgets.QWidget(self.centralwidget)
        self.widget_2.setMaximumSize(QtCore.QSize(800, 60))
        self.widget_2.setObjectName("widget_2")
        self.gridLayout = QtWidgets.QGridLayout(self.widget_2)
        self.gridLayout.setHorizontalSpacing(60)
        self.gridLayout.setObjectName("gridLayout")
        self.btn_annotate = QtWidgets.QPushButton(self.widget_2)
        self.btn_annotate.setMaximumSize(QtCore.QSize(120, 16777215))
        self.btn_annotate.setStyleSheet("background-color: rgb(60, 60, 80);")
        self.btn_annotate.setObjectName("btn_annotate")
        self.gridLayout.addWidget(self.btn_annotate, 0, 0, 1, 1)
        self.btn_capture = QtWidgets.QPushButton(self.widget_2)
        self.btn_capture.setMaximumSize(QtCore.QSize(120, 16777215))
        self.btn_capture.setStyleSheet("background-color: rgb(60, 70, 80);")
        self.btn_capture.setObjectName("btn_capture")
        self.gridLayout.addWidget(self.btn_capture, 0, 2, 1, 1)
        self.btn_switch_webcam = QtWidgets.QPushButton(self.widget_2)
        self.btn_switch_webcam.setMaximumSize(QtCore.QSize(120, 16777215))
        self.btn_switch_webcam.setStyleSheet("background-color: rgb(70, 60, 80);")
        self.btn_switch_webcam.setObjectName("btn_switch_webcam")
        self.gridLayout.addWidget(self.btn_switch_webcam, 0, 1, 1, 1)
        self.gridLayout_2.addWidget(self.widget_2, 0, 2, 1, 1)
        self.Label_Categories = QtWidgets.QLabel(self.centralwidget)
        self.Label_Categories.setMaximumSize(QtCore.QSize(16777215, 30))
        self.Label_Categories.setAlignment(QtCore.Qt.AlignCenter)
        self.Label_Categories.setObjectName("Label_Categories")
        self.gridLayout_2.addWidget(self.Label_Categories, 1, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1536, 22))
        self.menubar.setObjectName("menubar")
        self.menuFIle = QtWidgets.QMenu(self.menubar)
        self.menuFIle.setObjectName("menuFIle")
        self.menuTutorials = QtWidgets.QMenu(self.menubar)
        self.menuTutorials.setObjectName("menuTutorials")
        self.menuContact = QtWidgets.QMenu(self.menubar)
        self.menuContact.setObjectName("menuContact")
        # Add the action to the menu
        # Create an action within the menu
        self.contact_action = QtWidgets.QAction("Author's Website", self.menubar)
        self.menuContact.addAction(self.contact_action)
        self.contact_action.triggered.connect(lambda: os.startfile("https://urielgarcilazo.com/index.html"))
        # Create an action within the menu
        self.tutorials_action = QtWidgets.QAction("Project Website", self.menubar)
        self.menuTutorials.addAction(self.tutorials_action)
        self.tutorials_action.triggered.connect(lambda: os.startfile("https://ugarcil.github.io/LivePyxel/"))
        
        # ##############################################
        # Create a container widget for the new buttons
        self.button_container = QtWidgets.QWidget(self.widget_2)
        self.button_container.setMaximumSize(QtCore.QSize(80, 16777215))  # Max width of 80
        self.button_container.setObjectName("button_container")

        # Create a vertical layout for the container
        self.vertical_layout = QtWidgets.QVBoxLayout(self.button_container)
        self.vertical_layout.setContentsMargins(0, 0, 0, 0)  # Remove margins if desired
        self.vertical_layout.setSpacing(2)  # Adjust spacing between elements

        # Create a horizontal layout for the buttons
        self.button_row = QtWidgets.QHBoxLayout()
        self.button_row.setContentsMargins(0, 0, 0, 0)
        self.button_row.setSpacing(2)

        # Create and add the first button
        self.prev_btn = QtWidgets.QPushButton(self.button_container)
        self.prev_btn.setMaximumSize(QtCore.QSize(40, 40))
        self.prev_btn.setObjectName("prev_btn")
        self.prev_btn.setText("")        
        icon_binmask = QtGui.QIcon()
        icon_binmask.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/prev_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.prev_btn.setIcon(icon_binmask)
        self.prev_btn.setIconSize(QtCore.QSize(20, 20))
        self.button_row.addWidget(self.prev_btn)

        # Create and add the second button
        self.next_btn = QtWidgets.QPushButton(self.button_container)
        self.next_btn.setMaximumSize(QtCore.QSize(40, 40))
        self.next_btn.setObjectName("next_btn")
        self.next_btn.setText("")        
        icon_binmask = QtGui.QIcon()
        icon_binmask.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/next_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.next_btn.setIcon(icon_binmask)
        self.next_btn.setIconSize(QtCore.QSize(20, 20))
        self.button_row.addWidget(self.next_btn)

        # Add the button row to the vertical layout
        self.vertical_layout.addLayout(self.button_row)

        # Create and add the number input field
        self.number_input = QtWidgets.QLineEdit(self.button_container)
        self.number_input.setMaximumSize(QtCore.QSize(80, 25))  # Width matches container
        self.number_input.setAlignment(QtCore.Qt.AlignCenter)
        self.number_input.setValidator(QtGui.QIntValidator())  # Only allow integers
        self.number_input.setObjectName("number_input")
        self.number_input.setPlaceholderText("0")  # Default placeholder text
        self.vertical_layout.addWidget(self.number_input)

        # Add the container widget to the main grid layout
        self.gridLayout.addWidget(self.button_container, 0, 3, 1, 1)
        
        # #####################################################
        self.btn_switch_webcam = QtWidgets.QPushButton(self.widget_2)
        self.btn_switch_webcam.setMaximumSize(QtCore.QSize(120, 16777215))
        self.btn_switch_webcam.setStyleSheet("background-color: rgb(70, 60, 80);")
        self.btn_switch_webcam.setObjectName("btn_switch_webcam")
        self.gridLayout.addWidget(self.btn_switch_webcam, 0, 1, 1, 1)
        self.gridLayout_2.addWidget(self.widget_2, 0, 2, 1, 1)

        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionNew = QtWidgets.QAction(MainWindow)
        self.actionNew.setObjectName("actionNew")
        self.actionNew.triggered.connect(self.new_file_dialog)
        self.actionLoad_Folder_Ctrl_L = QtWidgets.QAction(MainWindow)
        self.actionLoad_Folder_Ctrl_L.setObjectName("actionLoad_Folder_Ctrl_L")
        self.actionLoad_Folder_Ctrl_L.triggered.connect(self.load_file_dialog)
        self.actionSaveFile = QtWidgets.QAction(MainWindow)
        self.actionSaveFile.setObjectName("actionSaveFile")
        self.actionSaveFile.triggered.connect(self.save_json_file)
        self.menuFIle.addAction(self.actionNew)
        self.menuFIle.addAction(self.actionLoad_Folder_Ctrl_L)
        self.menuFIle.addAction(self.actionSaveFile)
        self.menubar.addAction(self.menuFIle.menuAction())
        self.menubar.addAction(self.menuTutorials.menuAction())
        self.menubar.addAction(self.menuContact.menuAction())
        
        # Paint tool buttons
        self.widget = QtWidgets.QWidget(self.centralwidget)
        self.widget.setMaximumSize(QtCore.QSize(500, 60))
        self.widget.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.widget.setAutoFillBackground(False)
        self.widget.setObjectName("widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout_2.setContentsMargins(9, -1, 9, -1)
        self.horizontalLayout_2.setSpacing(8)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.polygon_button = QtWidgets.QPushButton(self.widget)
        self.polygon_button.setMaximumSize(QtCore.QSize(40, 40))
        self.polygon_button.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/polygon_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.polygon_button.setIcon(icon1)
        self.polygon_button.setIconSize(QtCore.QSize(30, 30))
        self.polygon_button.setObjectName("polygon_button")
        self.horizontalLayout_2.addWidget(self.polygon_button)
        self.brush_button = QtWidgets.QPushButton(self.widget)
        self.brush_button.setMaximumSize(QtCore.QSize(40, 40))
        self.brush_button.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/brush_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.brush_button.setIcon(icon2)
        self.brush_button.setIconSize(QtCore.QSize(30, 30))
        self.brush_button.setObjectName("brush_button")
        self.horizontalLayout_2.addWidget(self.brush_button)
        # Bezier button
        self.bezier_button = QtWidgets.QPushButton(self.widget)
        self.bezier_button.setMaximumSize(QtCore.QSize(40, 40))
        self.bezier_button.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/bezier_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.bezier_button.setIcon(icon2)
        self.bezier_button.setIconSize(QtCore.QSize(30, 30))
        self.bezier_button.setObjectName("bezier_button")
        self.horizontalLayout_2.addWidget(self.bezier_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # isAdditive button
        self.isAdditive_button = QtWidgets.QPushButton(self.widget)
        self.isAdditive_button.setMaximumSize(QtCore.QSize(40, 40))
        self.isAdditive_button.setText("")        
        self.isAdditive_button_on_icon = QtGui.QIcon()
        self.isAdditive_button_on_icon.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/additive_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.isAdditive_button_off_icon = QtGui.QIcon()
        self.isAdditive_button_off_icon.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/additiveOFF_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.isAdditive_button.setIcon(self.isAdditive_button_on_icon)
        self.isAdditive_button.setIconSize(QtCore.QSize(30, 30))
        self.isAdditive_button.setObjectName("isAdditive_button")
        self.horizontalLayout_2.addWidget(self.isAdditive_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # isTopLayerOnly button
        self.isTopLayerOnly_button = QtWidgets.QPushButton(self.widget)
        self.isTopLayerOnly_button.setMaximumSize(QtCore.QSize(40, 40))
        self.isTopLayerOnly_button.setText("")        
        self.isTopLayerOnly_button_on_icon = QtGui.QIcon()
        self.isTopLayerOnly_button_on_icon.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/topmask_ON_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.isTopLayerOnly_button_off_icon = QtGui.QIcon()
        self.isTopLayerOnly_button_off_icon.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/topmask_OFF_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.isTopLayerOnly_button.setIcon(self.isTopLayerOnly_button_on_icon)
        self.isTopLayerOnly_button.setIconSize(QtCore.QSize(30, 30))
        self.isTopLayerOnly_button.setObjectName("isTopLayerOnly_button")
        self.horizontalLayout_2.addWidget(self.isTopLayerOnly_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # # Bucket button
        self.bucket_button = QtWidgets.QPushButton(self.widget)
        self.bucket_button.setMaximumSize(QtCore.QSize(40, 40))
        self.bucket_button.setText("")        
        icon_bucket = QtGui.QIcon()
        icon_bucket.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/bucket_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.bucket_button.setIcon(icon_bucket)
        self.bucket_button.setIconSize(QtCore.QSize(30, 30))
        self.bucket_button.setObjectName("bucket_button")
        self.horizontalLayout_2.addWidget(self.bucket_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # # binmask button
        self.binmask_button = QtWidgets.QPushButton(self.widget)
        self.binmask_button.setMaximumSize(QtCore.QSize(40, 40))
        self.binmask_button.setText("")        
        icon_binmask = QtGui.QIcon()
        icon_binmask.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/binmask_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.binmask_button.setIcon(icon_binmask)
        self.binmask_button.setIconSize(QtCore.QSize(30, 30))
        self.binmask_button.setObjectName("binmask_button")
        self.horizontalLayout_2.addWidget(self.binmask_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # # OpenFolder button
        self.openFolder_button = QtWidgets.QPushButton(self.widget)
        self.openFolder_button.setMaximumSize(QtCore.QSize(40, 40))
        self.openFolder_button.setText("")        
        icon_bucket = QtGui.QIcon()
        icon_bucket.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/openFolder_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.openFolder_button.setIcon(icon_bucket)
        self.openFolder_button.setIconSize(QtCore.QSize(30, 30))
        self.openFolder_button.setObjectName("openFolder_button")
        self.horizontalLayout_2.addWidget(self.openFolder_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # # Freehand button
        self.freehand_button = QtWidgets.QPushButton(self.widget)
        self.freehand_button.setMaximumSize(QtCore.QSize(40, 40))
        self.freehand_button.setText("")        
        icon_bucket = QtGui.QIcon()
        icon_bucket.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/freehand_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.freehand_button.setIcon(icon_bucket)
        self.freehand_button.setIconSize(QtCore.QSize(30, 30))
        self.freehand_button.setObjectName("freehand_button")
        self.horizontalLayout_2.addWidget(self.freehand_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # Edit images button
        self.editMode_button = QtWidgets.QPushButton(self.widget)
        self.editMode_button.setMaximumSize(QtCore.QSize(40, 40))
        self.editMode_button.setText("")        
        icon_binmask = QtGui.QIcon()
        icon_binmask.addPixmap(QtGui.QPixmap(jn(self.script_path,"icons/editFolder_icon.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.editMode_button.setIcon(icon_binmask)
        self.editMode_button.setIconSize(QtCore.QSize(30, 30))
        self.editMode_button.setObjectName("editMode_button")
        self.horizontalLayout_2.addWidget(self.editMode_button)
        self.gridLayout_2.addWidget(self.widget, 1, 2, 1, 1)
        # 
        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        
        
        self.lobutton_label_blocks = [self.label_block_0]
        brush_settings["color"] = [0, 255, 0, 0]

        
    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "LivePyxel"))
        self.label_select_0.setText(_translate("MainWindow", "TextLabel"))
        self.label_6.setText(_translate("MainWindow", "Mask opacity"))
        self.btn_add.setText(_translate("MainWindow", "Add +"))
        self.btn_annotate.setText(_translate("MainWindow", "Annotate"))
        self.btn_capture.setText(_translate("MainWindow", "Capture \n"
"(Spacebar)"))
        self.btn_capture.setShortcut(_translate("MainWindow", "Space"))
        self.btn_switch_webcam.setText(_translate("MainWindow", "Switch Webcam\n"
"(Ctrl+1)"))
        self.btn_switch_webcam.setShortcut(_translate("MainWindow", "Ctrl+1"))
        self.Label_Categories.setText(_translate("MainWindow", "Label Categories"))
        self.menuFIle.setTitle(_translate("MainWindow", "FIle"))
        self.menuTutorials.setTitle(_translate("MainWindow", "Tutorials"))
        self.menuContact.setTitle(_translate("MainWindow", "Contact"))
        self.actionNew.setText(_translate("MainWindow", "New... (Ctrl+N)"))
        self.actionNew.setShortcut(_translate("MainWindow", "Ctrl+N"))
        self.actionLoad_Folder_Ctrl_L.setText(_translate("MainWindow", "Load Folder (Ctrl+L)"))
        self.actionLoad_Folder_Ctrl_L.setShortcut(_translate("MainWindow", "Ctrl+L"))
        self.actionSaveFile.setText(_translate("MainWindow", "Save File (Ctrl+S)"))
        self.actionSaveFile.setShortcut(_translate("MainWindow", "Ctrl+S"))

    def add_btn_block(self, label_text="New Label", color=None):
        # get the names of the color, label and edit buttons based on the value of index
        suffix = str(len(self.lobutton_label_blocks))
        
        
        # Create a new widget (btn_block)
        btn_block = QtWidgets.QWidget(self.scrollAreaWidgetContents_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        btn_block.setSizePolicy(sizePolicy)
        btn_block.setMaximumSize(QtCore.QSize(16777215, 40))

        # Create a horizontal layout for the new block
        horizontal_layout = QtWidgets.QHBoxLayout(btn_block)

        # Create and add the label
        label = QtWidgets.QLabel(btn_block)
        label.setObjectName(f"label_select_{suffix}")
        label.setText(label_text)
        horizontal_layout.addWidget(label)

        # Create and add the color button
        color_button = QtWidgets.QPushButton(btn_block)
        color_button.setMaximumSize(QtCore.QSize(60, 16777215))
        if color is None:
                new_color = ",".join([str(random.randint(125,255)) for _ in range(3)])
                color_button.setStyleSheet(f"background-color: rgb({new_color});")
        else:
                color = ",".join(str(c) for c in color)
                color_button.setStyleSheet(f"background-color: rgb({color});")
        color_button.setObjectName(f"btn_select_{suffix}")
        horizontal_layout.addWidget(color_button)
        
        ############# EDIT BUTTON PROCEDURALLY GENERATED ###################
        # Create and add the edit button with an icon
        edit_button = QtWidgets.QPushButton(btn_block)
        edit_button.setMinimumSize(QtCore.QSize(20, 0))
        edit_button.setMaximumSize(QtCore.QSize(30, 16777215))
        # icon = QtGui.QIcon()
        # icon.addPixmap(QtGui.QPixmap(self.icon), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        edit_button.setIcon(self.icon_edit)
        edit_button.setObjectName(f"btn_edit_{suffix}")
        horizontal_layout.addWidget(edit_button)
        
        ############# TRASH BUTTON PROCEDURALLY GENERATED ###################
        # Create and add the trash button with an icon
        trash_button = QtWidgets.QPushButton(btn_block)
        trash_button.setMinimumSize(QtCore.QSize(20, 0))
        trash_button.setMaximumSize(QtCore.QSize(30, 16777215))
        # icon = QtGui.QIcon()
        # icon.addPixmap(QtGui.QPixmap(self.icon), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        trash_button.setIcon(self.icon_trash)
        trash_button.setObjectName(f"btn_remove_{suffix}")
        horizontal_layout.addWidget(trash_button)
        
        

        # Find the number of existing blocks (rows) in gridLayout_6 to determine the next available row
        row_position = self.gridLayout_6.rowCount()

        # Add the new block widget to the gridLayout_6 at the next available row
        self.gridLayout_6.addWidget(btn_block, row_position, 0, 1, 1)
        
        self.lobutton_label_blocks.append(btn_block)
        edit_button.clicked.connect(lambda: self.traverse_edit_buttons(edit_button))
        color_button.clicked.connect(lambda: self.traverse_select_buttons(color_button))
        trash_button.clicked.connect(lambda: self.traverse_remove_buttons(color_button))
        
        
    def traverse_remove_buttons(self,button):
        parent_widget = button.parent()
        for idx,w in enumerate(self.lobutton_label_blocks):
            if w == parent_widget:
                # print(label.objectName(),color_edit.objectName())
                
                dialog = QtWidgets.QDialog()
                ui_dialog = Ui_Dialog_Delete()
                ui_dialog.setupUi(dialog, idx, self.lobutton_label_blocks, button)
                # Execute the dialog and check if accepted or rejected
                result = dialog.exec_()
                # open the new edit widget and pass the elements

    # use the identity of the buttons edit pressed (edit buttons are located next to the color buttons in the label categories widget) 
    # to get parent information, to extract label and color buttons
    def traverse_edit_buttons(self, button):
        parent_widget = button.parent()
        for idx,w in enumerate(self.lobutton_label_blocks):
            if w == parent_widget:
                # extract the color button and the label
                label = parent_widget.findChild(QtWidgets.QLabel, f"label_select_{idx}")
                color_edit = parent_widget.findChild(QtWidgets.QPushButton, f"btn_select_{idx}")
                # print(label.objectName(),color_edit.objectName())
                
                dialog = QtWidgets.QDialog()
                ui_dialog = Ui_Dialog()
                ui_dialog.setupUi(dialog, label, color_edit)
                # Execute the dialog and check if accepted or rejected
                result = dialog.exec_()
                # open the new edit widget and pass the elements
                
        
    def traverse_select_buttons(self,button):
        global brushColor
        buttonStyle = button.styleSheet()
        buttonStyle = buttonStyle.replace(" ","")
        if "background-color:rgb" in buttonStyle:
            # buttonStyle = buttonStyle.replace(" ","") #remove empty spaces
            string_color = buttonStyle.split("background-color:rgb")[1] #extract everything after background-color
            string_color = string_color.split(";")[0] # keep everything before the end of the syle, defined by ;
            # turn digits into a tuple
            string_color = string_color[1:-1] #remove parenthesis
            string_color = string_color.split(",")
            colors = [int(c) for c in string_color]
            string_color = [colors[2],colors[1],colors[0]]
            # print(string_color)
            brush_settings["color"] = tuple(string_color) + (0,)
            # print(brush_settings["color"])
    def createSubfolders(self):
        os_settings["masks_path"] = jn(os_settings["directory"],"masks")
        os_settings["images_path"] = jn(os_settings["directory"],"images")
        if not os.path.exists(os_settings["masks_path"]):
            os.mkdir(os_settings["masks_path"])
        if not os.path.exists(os_settings["images_path"]):
            os.mkdir(os_settings["images_path"])
        
        
        
    
    def save_json_file(self):
        def get_color(stylesheet:str):
            stylesheet = stylesheet.replace(" ","")
            stylesheet = stylesheet.split("background-color:rgb")[1]
            stylesheet = stylesheet.split(";")[0]
            colors = stylesheet[1:-1].split(",")
            colors = [int(c) for c in colors]
            # turn it into bgr, which is cv2 default
            # colors = (colors[0],colors[1],colors[2])
            return colors
        
        
        if os_settings["config"] != "":
            labels = []
            colors = []
            for idx,block in enumerate(self.lobutton_label_blocks):          
                label = block.findChild(QtWidgets.QLabel, f"label_select_{idx}")
                labels.append(label.text())
                color_edit = block.findChild(QtWidgets.QPushButton, f"btn_select_{idx}")
                colors.append(get_color(color_edit.styleSheet()))
                
            config = {
                "labels":labels,
                "colors":colors
            }
            
            with open(os_settings["config"],"w") as jsonfile:
                json.dump(config, jsonfile, indent=4)  # Use json.dump to write the dictionary to the file
            QtWidgets.QMessageBox.information(None, "Success!!!", f"The project has been saved")
        else:
            QtWidgets.QMessageBox.critical(None,"Project couldn't be saved", "It seems no config file has been selected.\nThis is probably caused by a missing path to the config.json file.\nPlease create or load a project first.")
            
    def new_file_dialog(self):
        options = QtWidgets.QFileDialog.Options()
        directory = QtWidgets.QFileDialog.getExistingDirectory(None, "Select Folder", "", options=options)
        if directory:
            # !!! system shouldn't be set before asking user to work around config.json file (i.e. no os_settings)
            os_settings["directory"] = directory
            os_settings["config"] = jn(directory, "config.json")
            
            if os.path.exists(os_settings["config"]):
                QtWidgets.QMessageBox.information(None, "A config file was found in the folder selected", f"A config.json file seems to already exist in the location \n\n{os_settings['config']}.\n\nPlease make sure to erase all data if you want to use this location, and try again")
            else:
                self.save_json_file()
                self.createSubfolders()
                QtWidgets.QMessageBox.information(None, "Success!!", f"A new project project has been created in the directory: {directory}")
                    # You can add additional code here to handle the new file creation        
        
        
    def clear_layout(self):
        for i in reversed(range(self.gridLayout_6.count())): 
            widget = self.gridLayout_6.itemAt(i).widget()
            if widget is not None: 
                widget.deleteLater()  # Properly delete the widget
            self.gridLayout_6.removeItem(self.gridLayout_6.itemAt(i))
        self.lobutton_label_blocks = []
                
    def load_file_dialog(self):
        options = QtWidgets.QFileDialog.Options()
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(
            None, 
            "Select JSON File", 
            "", 
            "JSON Files (*.json);;All Files (*)", 
            options=options
        )
        
        if file_name:
            if os.path.exists(file_name):
                with open(file_name, "r") as jsonfile:
                    config = json.load(jsonfile)
                    labels = config.get("labels", [])
                    colors = config.get("colors", [])

                # cv2 understands colors as  BGR
                # bgr_colors = []
                # for color in colors:
                #     # Swap R and B channels (assuming color is [R,G,B])
                #     bgr_color = [color[2], color[1], color[0]]
                #     bgr_colors.append(bgr_color)
                    
                self.clear_layout()
                for idx in range(len(labels)):
                    self.add_btn_block(labels[idx],colors[idx])
                QtWidgets.QMessageBox.information(None, "Success!!", f"The information from the previous session has been successfully loaded")
                os_settings["config"] = file_name
                os_settings["masks_path"] = jn(os.path.split(file_name)[0],"masks")
                os_settings["images_path"] = jn(os.path.split(file_name)[0],"images")
                print(os_settings["images_path"])
                print(os_settings["masks_path"])

            else:
                print("Selected file not found.")
        else:
            print("No file selected.")

    
# DD. DISPLAY_WIDGET
# displayWidget = WebcamWidget(QLabel)
# interp. a child class optimized for event controlling
class WebcamWidget(QtWidgets.QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        custom_cursor_pixmap_polygon = QtGui.QPixmap("./icons/bright_cross_cursor.png")  # Ensure you have a bright cursor image
        self.custom_cursor_polygon = QtGui.QCursor(custom_cursor_pixmap_polygon)
        self.setMouseTracking(True)

    def enterEvent(self, event):
        """Hide the cursor when the mouse enters the widget."""
        if brush_settings["is_brush_mode"] == "brush":
            self.setCursor(Qt.BlankCursor)
        elif brush_settings["is_brush_mode"] == "polygon" or brush_settings["is_brush_mode"] == "bezier":
            self.setCursor(self.custom_cursor_polygon)
        cursor_settings["in_display"] = True

    def leaveEvent(self, event):
        """Restore the cursor when the mouse leaves the widget."""
        self.setCursor(Qt.ArrowCursor)
        cursor_settings["in_display"] = False

    
if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())


